/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlCudaMemHandler.cpp
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**  Handles NVIDIA Cuda memory interface.
**
*****************************************************************************/

#include "svlCudaMemHandler.h"

// Static members.
bool svlCudaMemHandler::track_allocations = false;
list<NamedLine> svlCudaMemHandler::nls;
list<NamedPitch> svlCudaMemHandler::nps;

// Default constructor
svlCudaMemHandler::svlCudaMemHandler()
{
}

svlCudaMemHandler::~svlCudaMemHandler()
{
}

// Send some random memory back and forth to the GPU, checking that it works (sanity check).
bool svlCudaMemHandler::testMemoryAccess(int w, int h)
{
	svlCudaPitch p;
	svlCudaAllocPitch(w, h, p);
	// Create random memory.
	float *data = (float*)malloc(w*h*sizeof(float));
	for ( int i=0; i<w*h; ++i )
		data[i] = float(i);
	// Copy to and back out in verify.
	svlCudaPitchToGPU(p, data);
	float *verify = (float*)malloc(w*h*sizeof(float));
	svlCudaPitchFromGPU(verify, p);
	// Verify that verify matches data.
	bool ret = true;
	for ( int i=0; i<w*h && ret; ++i ) {
		if ( data[i] != verify[i] ) {
			printf("Mismatch at %d: %f %f\n", i, data[i], verify[i]);
			ret = false;
		}
	}
	svlCudaFreePitch(p);
	free(data);
	free(verify);
	return ret;
}

// Write a pitch out to file.
void svlCudaMemHandler::pitch2file(const char *file, svlCudaPitch *p)
{
	float *back = (float*)malloc(sizeof(float)*p->size());
	pitchFromGPU(back, p);
	FILE *fo = fopen(file,"wb");
	fwrite(&p->w,sizeof(int),1,fo);
	fwrite(&p->h,sizeof(int),1,fo);
	fwrite(back,sizeof(float),p->w*p->h,fo);
	fclose(fo);
	free(back);
}

// Write a line out to file.
void svlCudaMemHandler::line2file(const char *file, svlCudaLine *n)
{
	float *back = (float*)malloc(sizeof(float)*n->w);
	lineFromGPU(back, n);
	FILE *fo = fopen(file,"wb");
	//fwrite(&n->w,sizeof(int),1,fo);
	fwrite(back,sizeof(float),n->w,fo);
	fclose(fo);
	free(back);
}

// Create an image containing the random indices of locations for fast cuda memory transfer and access on the device.
int svlCudaMemHandler::fillIndexImage(const vector<CvPoint> &locations, IplImage *ixage)
{
#if 0
	// Pointers for filling in each row.
	vector<int> col_ptrs(ixage->height, 0);
	// Fill in index image.
	for ( vector<CvPoint>::const_iterator it=locations.begin(); it!=locations.end(); ++it ) {
		CV_IMAGE_ELEM(ixage, int, it->y, col_ptrs[it->y]++) = it->x;
	}
	// Find max population in each row.
	int max_gi = col_ptrs[0];
	for ( vector<int>::const_iterator it=col_ptrs.begin(); it!=col_ptrs.end(); ++it )
		if ( *it > max_gi )
			max_gi = *it;
	// -1-alize remaining entries up to align-by-sixteen.
	int max_col = iAlignUp(max_gi, 16);
	for ( int r=0; r<ixage->height; ++r ) {
		for ( int c=col_ptrs[r]; c<max_col; ++c ) {
			CV_IMAGE_ELEM(ixage, int, r, c) = -1;
		}
	}
#else
	// Collaps row-column indices into image linearly.
	int ix = 0, r = 0, c = 0;
	for ( vector<CvPoint>::const_iterator loc=locations.begin(); loc!=locations.end(); ++loc ) {
		CV_IMAGE_ELEM(ixage, int, r, c++) = loc->y; // Row
		CV_IMAGE_ELEM(ixage, int, r, c++) = loc->x; // Column
		if ( c == ixage->width ) {
			++r;
			c = 0;
		}
	}
	if ( c == 0 ) {
		--r;
		c = ixage->width;
	}
	// -1-alize last row;
	for ( ; c<ixage->width; ++c ) {
		CV_IMAGE_ELEM(ixage, int, r, c) = -1;
	}
	return r+1;
#endif
}

// Fill a line with the random indices of locations for fast cuda memory transfer and access on the device.
void svlCudaMemHandler::fillIndexLine(const vector<CvPoint> &locations, svlCudaLine *ixline)
{
	// Collaps row-column indices into line linearly.
	int *data = (int*)malloc(sizeof(int)*int(locations.size())*2);
	int *ptr = data;
	for ( vector<CvPoint>::const_iterator loc=locations.begin(); loc!=locations.end(); ++loc ) {
		*ptr++ = loc->y; // Row
		*ptr++ = loc->x; // Column
	}
	lineToGPU(ixline, (float*)data, int(locations.size())*2);
	free(data);
}

// Print total amount of allocation.
void svlCudaMemHandler::printTotalAllocation() {
	if ( !track_allocations ) {
		cout << "Allocations are not being tracked.\n";
		return;
	}
	int line_sz = 0, pitch_sz = 0;
	for ( list<NamedLine>::const_iterator it = nls.begin(); it != nls.end(); ++it )
		line_sz += it->first->w;
	for ( list<NamedPitch>::const_iterator it = nps.begin(); it != nps.end(); ++it )
		pitch_sz += it->first->w * it->first->h;
	printf("Total device allocation:\n");
	printf("  %d lines for %d * %d bytes\n", int(nls.size()), line_sz, sizeof(float));
	printf("  %d pitches for %d * %d bytes\n", int(nps.size()), pitch_sz, sizeof(float));
	printf("Total: %d structure for %d * %d bytes\n", int(nls.size() + nps.size()), line_sz + pitch_sz, sizeof(float));
}

// Print all individual allocations.
void svlCudaMemHandler::printAllAllocations() {
	if ( !track_allocations ) {
		cout << "Allocations are not being tracked.\n";
		return;
	}
	int total_sz = 0;
	printf("All device allocations:\n");
	for ( list<NamedLine>::const_iterator it = nls.begin(); it != nls.end(); ++it ) {
		if ( it->second == NULL ) {
			printf("  line @ %x: %d * %d bytes\n", it->first->data, it->first->w, sizeof(float));
		} else {
			printf("  line \"%s\" @ %x: %d * %d bytes\n", it->second, it->first->data, it->first->w, sizeof(float));
		}
		total_sz += it->first->w;
	}
	for ( list<NamedPitch>::const_iterator it = nps.begin(); it != nps.end(); ++it ) {
		if ( it->second == NULL ) {
			printf("  pitch @ %x: %d * %d bytes\n", it->first->data, it->first->w * it->first->h, sizeof(float));
		} else {
			printf("  pitch \"%s\" @ %x: %d * %d bytes\n", it->second, it->first->data, it->first->w * it->first->h, sizeof(float));
		}
		total_sz += it->first->w * it->first->h;
	}
	printf("Total device allocation: %d structures for %d * %d bytes\n", int(nls.size() + nps.size()), total_sz, sizeof(float));
}

// Specialization for CvPoint.
void svlCudaMemHandler::CvPoint2IntLine(const vector<CvPoint> &vec, svlCudaLine *line, bool yx)
{
	// Collaps row-column indices into line linearly.
	int *data = (int*)malloc(sizeof(int)*2*int(vec.size()));
	int *ptr = data;
	for ( vector<CvPoint>::const_iterator v=vec.begin(); v!=vec.end(); ++v ) {
		if ( yx ) {
			*ptr++ = v->y;
			*ptr++ = v->x;
		} else {
			*ptr++ = v->x;
			*ptr++ = v->y;
		}
	}
	lineToGPU(line, (float*)data, 2*int(vec.size()));
	free(data);
}


