/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlCircularIterator.h
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**   Class for circularly iterating around vectors and lists, or any structure
**   with iterators and respective begin, end, rbegin, and rend.
*****************************************************************************/

#ifdef _WIN32

#pragma once

#include <stdio.h>
#include <iostream>

using namespace std;

template<class C, class V>
class svlCircularIterator {
private:
	typename C::iterator it;
	C *ptr;

public:
	inline svlCircularIterator() {
		ptr = NULL;
	}

	template<class C1> inline svlCircularIterator(C1 *list_) {
		ptr = (C*)list_;
		it = (typename C::iterator)list_->begin();
	}

	~svlCircularIterator() {}

	// Functions to manually set iterator; use = H.end(), et al., for great visibility.
	inline void begin();
	inline void front();
	inline void back();
	inline void end();

	// Getters and setters.
	inline typename C::iterator get_iter() const;
	inline void set_iter(const typename C::iterator& it_);
	inline C* get_ptr() const;
	inline void set_ptr(C* ptr_);

	// Assignment and equality operators.
	inline typename C::iterator& operator=(const typename C::iterator& it_);
	inline typename C::iterator& operator=(const typename C::reverse_iterator& it_);
	inline svlCircularIterator<C,V>& operator=(const svlCircularIterator<C,V>& it_);
	inline bool operator==(const typename C::iterator& it_) const;
	inline bool operator!=(const typename C::iterator& it_) const;
	inline bool operator<=(const typename C::iterator& it_) const;
	inline bool operator==(const svlCircularIterator<C,V>& cit) const;
	inline bool operator!=(const svlCircularIterator<C,V>& cit) const;
	inline bool operator<=(const svlCircularIterator<C,V>& cit) const;

	// Dereference, inc/decrement, and addition/subtraction operators.
	inline V& operator*() const;
	inline typename C::iterator& operator->();
	inline typename C::iterator& operator++();
	inline typename C::iterator& operator--();
	template<class V1> inline typename C::iterator& operator+=(V1 i);
	template<class V1> inline typename C::iterator& operator-=(V1 i);
	template<class V1> friend inline typename C::iterator operator+(const svlCircularIterator<C,V>& ci, V1 i);
	template<class V1> friend inline typename C::iterator operator-(const svlCircularIterator<C,V>& ci, V1 i);
};

template<class C, class V>
inline void svlCircularIterator<C,V>::begin() {
	if ( ptr ) {
		it = ptr->begin();
	}
}

template<class C, class V>
inline void svlCircularIterator<C,V>::front() {
	if ( ptr ) {
		it = ptr->begin();
	}
}

template<class C, class V>
inline void svlCircularIterator<C,V>::back() {
	if ( ptr ) {
		it = ptr->end();
		--it;
	}
}

template<class C, class V>
inline void svlCircularIterator<C,V>::end() {
	if ( ptr ) {
		it = ptr->end();
		--it;
	}
}

template<class C, class V>
inline typename C::iterator svlCircularIterator<C,V>::get_iter() const {
	return it;
}

template<class C, class V>
inline void svlCircularIterator<C,V>::set_iter(const typename C::iterator& it_) {
	it = it_;
}

template<class C, class V>
inline C* svlCircularIterator<C,V>::get_ptr() const {
	return ptr;
}

template<class C, class V>
inline void svlCircularIterator<C,V>::set_ptr(C* ptr_) {
	ptr = ptr_;
}

template<class C, class V>
inline typename C::iterator& svlCircularIterator<C,V>::operator=(const typename C::iterator& it_) {
	it = it_;
	if ( ptr && it == ptr->end() ) {
		it = ptr->begin();
	}
	return it;
}

template<class C, class V>
inline typename C::iterator& svlCircularIterator<C,V>::operator=(const typename C::reverse_iterator& it_) {
	it = it_.base();
	--it;
	if ( ptr && it == ptr->end() ) {
		it = ptr->begin();
	}
	return it;
}

template<class C, class V>
inline svlCircularIterator<C,V>& svlCircularIterator<C,V>::operator=(const svlCircularIterator<C,V>& it_) {
	ptr = it_.get_ptr();
	it = it_.get_iter();
	return *this;
}

template<class C, class V>
inline bool svlCircularIterator<C,V>::operator==(const typename C::iterator& it_) const {
	return it==it_;
}

template<class C, class V>
inline bool svlCircularIterator<C,V>::operator!=(const typename C::iterator& it_) const {
	return it!=it_;
}

template<class C, class V>
inline bool svlCircularIterator<C,V>::operator<=(const typename C::iterator& it_) const {
	return it<=it_;
}

template<class C, class V>
inline bool svlCircularIterator<C,V>::operator==(const svlCircularIterator<C,V>& cit) const {
	return it==cit.get_iter();
}

template<class C, class V>
inline bool svlCircularIterator<C,V>::operator!=(const svlCircularIterator<C,V>& cit) const {
	return it!=cit.get_iter();
}

template<class C, class V>
inline bool svlCircularIterator<C,V>::operator<=(const svlCircularIterator<C,V>& cit) const {
	return it<=cit.get_iter();
}

template<class C, class V>
inline V& svlCircularIterator<C,V>::operator*() const {
	return *it;
}

template<class C, class V>
inline typename C::iterator& svlCircularIterator<C,V>::operator->() {
	return this->it;
}

template<class C, class V>
inline typename C::iterator& svlCircularIterator<C,V>::operator++() {
	if ( ptr ) {
		++it;
		if ( it == ptr->end() )
			it = ptr->begin();
	}
	return this->it;
}

template<class C, class V>
inline typename C::iterator& svlCircularIterator<C,V>::operator--() {
	if ( ptr ) {
		if ( it == ptr->begin() ) {
			it = ptr->rbegin().base();
			--it;
		} else
			--it;
	}
	return this->it;
}

template<class C, class V>
template<class V1>
inline typename C::iterator& svlCircularIterator<C,V>::operator+=(V1 i) {
	int inti = abs(i);
	while ( --inti >= 0 )
		++*this;
	return this->it;
}

template<class C, class V>
template<class V1>
inline typename C::iterator& svlCircularIterator<C,V>::operator-=(V1 i) {
	int inti = abs(i);
	while ( --inti >= 0 )
		--*this;
	return this->it;
}

/****** Friend operators ******/

template<class C, class V, class V1>
inline typename C::iterator operator+(const svlCircularIterator<C,V>& ci, V1 i) {
	svlCircularIterator<C,V> copy = ci;
	int i_ = abs(i);
	while ( --i_ >= 0 )
		++copy;
	return copy.get_iter();
}

template<class C, class V, class V1>
inline typename C::iterator operator-(const svlCircularIterator<C,V>& ci, V1 i) {
	svlCircularIterator<C,V> copy = ci;
	int i_ = abs(i);
	while ( --i_ >= 0 )
		--copy;
	return copy.get_iter();
}

#endif

