/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlKdTree.cpp
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
**
** DESCRIPTION:
**   Source file for the kd-tree class.
*****************************************************************************/

#include "svlKdTree.h"

#define DIST_TOL 1e-4

// CONSTRUCTORS AND DESTRUCTORS

svlKdTree::svlKdTree() : _troot(NULL), _P(NULL), _bucket_size(1) {
}

svlKdTree::svlKdTree(vector<svlPoint3d> *P_in) : _troot(NULL), _bucket_size(1) {
	createKdTree(P_in);
}

svlKdTree::svlKdTree(vector<svlPoint3d> *P_in, int _bucket_size) : _troot(NULL), _bucket_size(_bucket_size) {
	createKdTree(P_in);
}

svlKdTree::svlKdTree(vector<svlPoint3d> *P_in, double min_box, vector<svlPoint3d> *P_out) : _troot(NULL), _bucket_size(1) {
	createKdTreeDec(P_in,min_box,P_out);
}

svlKdTree::svlKdTree(vector<svlPoint3d> *P_in, double min_box, vector<svlPoint3d> *P_out, vector<int> *in2out) : _troot(NULL), _bucket_size(1) {
	createKdTreeDec(P_in,min_box,P_out, in2out);
}

svlKdTree::svlKdTree(vector<svlPoint3d> *P_in, double min_box, vector<svlPoint3d> *P_out, int bucket_size, vector<int> *in2out) : _troot(NULL), _bucket_size(1) {
	createKdTreeDec(P_in,min_box,P_out,bucket_size, in2out);
}

svlKdTree::~svlKdTree() {
	// loop through all nodes and delete them
	if ( _troot != NULL ) {
		deleteNodes(_troot);
		/*if ( _troot->left != NULL )
			deleteNodes(_troot->left);
		if ( _troot->right != NULL )
			deleteNodes(_troot->right);
		_troot->ix.clear();
		delete _troot;*/
	}
	_P = NULL;
	_pts2nodes.clear();
	_max_dists.clear();
}

// METHODS

void svlKdTree::deleteNodes(svlTreeNode* node) {
	if ( node->left != NULL )
		deleteNodes(node->left);
	if ( node->right != NULL )
		deleteNodes(node->right);
	node->ix.clear();
	delete node;
}

void svlKdTree::connectParents(svlTreeNode *parent, svlTreeNode *node) {
	node->parent = parent;
	if ( node->left != NULL )
		connectParents(node,node->left);
	if ( node->right != NULL )
		connectParents(node,node->right);
}

void svlKdTree::createKdTree(vector<svlPoint3d> *P_in, int bucket_size) {
	_bucket_size = bucket_size;
	createKdTree(P_in);
}

void svlKdTree::createKdTree(vector<svlPoint3d> *P_in) {
	debug1 = false;

	if ( _troot != NULL ) {
		if ( _troot->left != NULL )
			deleteNodes(_troot->left);
		if ( _troot->right != NULL )
			deleteNodes(_troot->right);
		delete _troot;
	}
	_P = P_in; // save pointer to points
	n_pts = (int)_P->size();
	vector<int> ix(n_pts,0); // allocate indices
	for ( int i=1; i<n_pts; ++i ) {
		ix[i] = i;
	}
	_pts2nodes = vector<svlTreeNode*>(n_pts,(svlTreeNode*)NULL); // NULL pointers for all points
	_max_dists = vector<double>(n_pts,-1.0);

	if ( (int)ix.size() <= _bucket_size ) {
		// just put all indices into root
		_troot = new svlTreeNode(); // new NULL node
		_troot->ix.reserve(ix.size());
		_troot->box = svlBox3d();
		for ( vector<int>::const_iterator i=ix.begin(); i!=ix.end(); ++i ) {
			_troot->ix.push_back(*i);
			_pts2nodes[*i] = _troot;
			_troot->box.addToBox((*_P)[*i]);
		}
	} else { // splitting long the median
		int dim;
		double span;
		int piv = select_best_median(ix, 0, n_pts-1, dim, span); // first splitting/partition value

		// build tree
		_troot = new svlTreeNode(ix[piv],
							 dim,
							 (*_P)[ix[piv]][dim],
							 createKdSubTree(ix,0,piv-1), // continue partitioning sub-trees
							 createKdSubTree(ix,piv+1,n_pts-1));
		_pts2nodes[ix[piv]] = _troot;

		// back-connect parent pointers
		if ( _troot->left != NULL )
			connectParents(_troot,_troot->left);
		if ( _troot->right != NULL )
			connectParents(_troot,_troot->right);

		// fill bounding boxes
		fillBoundingBoxes(_troot);
	}
}

svlTreeNode* svlKdTree::createKdSubTree(vector<int> &ix, int left, int right) {
	if ( (right-left+1) <= _bucket_size ) { // just put all these into the current node
		svlTreeNode* ret = new svlTreeNode(); // leaf node
		ret->ix.reserve(right-left+1);
		for ( int i=left; i<=right; ++i ) {
			ret->ix.push_back(ix[i]);
			_pts2nodes[ix[i]] = ret;
		}
		return ret;
	}
	// else, we have to split the points up
	int dim;
	double span;
	int piv = select_best_median(ix, left, right, dim, span); // first splitting/partition value
	svlTreeNode *ret = new svlTreeNode(ix[piv],
								 dim,
								 (*_P)[ix[piv]][dim],
								 ( piv-1-left < 0 ? NULL : createKdSubTree(ix, left, piv-1) ),
								 ( right-piv-1 < 0 ? NULL : createKdSubTree(ix, piv+1, right) ));
	_pts2nodes[ix[piv]] = ret;
	return ret;
}

void svlKdTree::fillBoundingBoxes(svlTreeNode *q) {
	q->box = svlBox3d();
	for ( vector<int>::const_iterator i=q->ix.begin(); i!=q->ix.end(); ++i ) {
		q->box.addToBox((*_P)[*i]);
	}
	if ( q->left != NULL ) {
		fillBoundingBoxes(q->left);
		q->box.unionWithBox(q->left->box);
	}
	if ( q->right != NULL ) {
		fillBoundingBoxes(q->right);
		q->box.unionWithBox(q->right->box);
	}
}

void svlKdTree::createKdTreeDec(vector<svlPoint3d> *P_in, double min_box, vector<svlPoint3d> *P_out, int bucket_size, vector<int> *in2out) {
	_bucket_size = 1;
	if ( in2out != NULL ) {
		in2out->clear();
		in2out->resize(P_in->size(), -1);
	}
	createKdTreeDec(P_in, min_box, P_out, in2out);
	rebucket(bucket_size);
}

void svlKdTree::createKdTreeDec(vector<svlPoint3d> *P_in, double min_box, vector<svlPoint3d> *P_out, vector<int> *in2out) {
	_bucket_size = 1;
	debug1 = false;

	if ( _troot != NULL ) {
		if ( _troot->left != NULL )
			deleteNodes(_troot->left);
		if ( _troot->right != NULL )
			deleteNodes(_troot->right);
		delete _troot;
	}
	_P = P_in; // save pointer to points
	n_pts = (int)_P->size();
	vector<int> ix(n_pts,0); // allocate indices
	for ( int i=1; i<n_pts; ++i ) {
		ix[i] = i;
	}
	// do not init _pts2nodes or _max_dists with values
#if KDT_DEBUG_OUT
	cout << _pts2nodes.size() << endl;
#endif

	int dim = 0;
	double span;
	int piv = select_best_median(ix, 0, n_pts-1, dim, span); // first splitting/partition value
	if ( span < min_box ) { // already in trouble: too populated
		// replace entire point cloud with one point
		svlPoint3d p(0.0);
		int k = 0;
		for ( vector<svlPoint3d>::const_iterator i=(*_P).begin(); i!=(*_P).end(); ++i ) {
			p += *i;
			if ( in2out != NULL )
				(*in2out)[k++] = 0;
		}
		p /= (double)n_pts;
		_troot = new svlTreeNode(0);
		P_out->push_back(p);
		_pts2nodes.push_back(_troot);
	} else { // split
		int k = ix[piv];
		if ( in2out != NULL )
			(*in2out)[k] = (int)P_out->size();
		P_out->push_back((*_P)[k]); // save split point to output cloud
		// build tree
		_troot = new svlTreeNode((int)(*P_out).size()-1, dim, (*P_out).back()[dim]);
		_pts2nodes.push_back(_troot);
		_troot->left = createKdSubTreeDec(ix,0,piv-1,min_box,P_out, in2out);
		_troot->right = createKdSubTreeDec(ix,piv+1,n_pts-1,min_box,P_out, in2out); // continue partitioning sub-trees
#if KDT_DEBUG_OUT
		cout << _pts2nodes.size() << ", " << _pts2nodes[0] << endl;
#endif

		// back-connect parent pointers
		if ( _troot->left != NULL )
			connectParents(_troot,_troot->left);
		if ( _troot->right != NULL )
			connectParents(_troot,_troot->right);
	}
	_P = P_out; // Redirect pointer to real cloud
	fillBoundingBoxes(_troot); // fill bounding boxes
	n_pts = (int)(*_P).size();
	_max_dists = vector<double>(n_pts,-1.0); // now init _max_dists
}

svlTreeNode* svlKdTree::createKdSubTreeDec(vector<int> &ix, int left, int right, double min_box, vector<svlPoint3d> *P_out, vector<int> *in2out) {
	int dim; // dim modified by reference in select_best_median
	double span;
	int piv = select_best_median(ix, left, right, dim, span); // first splitting/partition value
	svlTreeNode *ret;
	if ( span < min_box ) { // replace these points with the average
		// replace entire point cloud with one point
		svlPoint3d p(0.0);
		for ( int i=left; i<=right; ++i ) {
			int k = ix[i];
			p += (*_P)[k];
			if ( in2out != NULL )
				(*in2out)[k] = (int)P_out->size();
		}
		p /= (double)(right-left+1);
		P_out->push_back(p);
		ret = new svlTreeNode((int)(*P_out).size()-1);
		_pts2nodes.push_back(ret);
	} else { // split
		int k = ix[piv];
		if ( in2out != NULL )
			(*in2out)[k] = (int)P_out->size();
		P_out->push_back((*_P)[k]); // save split point to output cloud
		ret = new svlTreeNode((int)(*P_out).size()-1, dim, (*P_out).back()[dim]);
		_pts2nodes.push_back(ret);
		ret->left = piv-1-left < 0 ? NULL : createKdSubTreeDec(ix, left, piv-1, min_box, P_out, in2out);
		ret->right = right-piv-1 < 0 ? NULL : createKdSubTreeDec(ix, piv+1, right, min_box, P_out, in2out);
	}
	return ret;
}

int svlKdTree::partition(vector<int> &ix, int left, int right, int piv, int dim) {
	// Partition based on a pivot.
	double pivot_val = (*_P)[ix[piv]].getD(dim);
	int temp = ix[piv];
	ix[piv] = ix[right];
	ix[right] = temp; // swap pivot out to right side
	int stix = left; // storage index
	for ( int i=left; i<right; ++i ) {
		if ( (*_P)[ix[i]][dim] < pivot_val ) {
			temp = ix[stix];
			ix[stix] = ix[i];
			ix[i] = temp;
			++stix;
		}
	}
	temp = ix[stix];
	ix[stix] = ix[right];
	ix[right] = temp;
	return stix;
}

int svlKdTree::select_median(vector<int> &ix, int left, int right, int med, int dim) {
	// Calculate median along dimension dim, capturing the min and max values along the way.
	int piv;
	for (;;) {
		piv = partition(ix, left, right, ((right-left)>>1)+left, dim);
		if ( piv == med )
			return piv;
		else if ( piv > med )
			right = piv-1;
		else
			left = piv+1;
	}
}

int svlKdTree::select_best_median(vector<int> &ix, int left, int right, int &dim, double &span) {
	svlPoint3d min_pt = (*_P)[ix[left]], max_pt = (*_P)[ix[left]];
	for ( int i=left+1; i<=right; ++i ) {
		min_pt.min_with((*_P)[ix[i]]);
		max_pt.max_with((*_P)[ix[i]]);
	}
	max_pt -= min_pt;
	span = max_pt.pt_max(); // save the max separation
	if ( max_pt.x > max_pt.y )
		if ( max_pt.x > max_pt.z )
			dim = 0;
		else
			dim = 2;
	else
		if ( max_pt.y > max_pt.z )
			dim = 1;
		else
			dim = 2;
	return select_median(ix, left, right, ((right-left)>>1)+left, dim);
}

void svlKdTree::rebucket(int bucket_size) {
	debug1 = false;
#if KDT_DEBUG_OUT
	cout << "Perhaps rebucketing ...\n";
#endif
	if ( _bucket_size != bucket_size ) {
#if KDT_DEBUG_OUT
		cout << "Rebucketing!\n";
#endif
		if ( _bucket_size > bucket_size ) { // we have to decrease our bucket size
			_bucket_size = bucket_size;
			// loop over all nodes
			int dim, piv;
			double span;
			svlTreeNode *q;
			for ( int i=0; i<n_pts; ++i ) {
				q = _pts2nodes[i];
				if ( (int)q->ix.size() > _bucket_size ) {
					vector<int> ix(q->ix); // copy indices
					piv = select_best_median(ix, 0, (int)ix.size()-1, dim, span); // get new best split
					q->ix.clear(); // define new node
					q->ix.push_back(ix[piv]);
					q->dim = dim;
					q->splitVal = (*_P)[ix[piv]][dim];
					q->left = piv-1 < 0 ? NULL : createKdSubTree(ix, 0, piv-1); // create new sub-trees
					q->right = (int)ix.size()-piv-2 < 0 ? NULL : createKdSubTree(ix, piv+1, (int)ix.size()-1);
					// back-connect parent pointers
					if ( q->left != NULL )
						connectParents(q,q->left);
					if ( q->right != NULL )
						connectParents(q,q->right);
					fillBoundingBoxes(q); // create bounding for sub nodes
				}
			}
		} else { // we need to collect buckets
			_bucket_size = bucket_size;
			svlTreeNode *q;
			int changed = 1;
			while ( changed > 0 ) {
				changed = 0;
				for ( int i=0; i<n_pts; ++i ) {
					q = _pts2nodes[i];
					if ( isLeaf(q) && (int)q->ix.size() < _bucket_size ) { // this node is possibly under-filled
						svlTreeNode* s = getSibling(q);
						if ( s == NULL // no nodes to add from sibling, so size+1 is <= new bucket size
							|| ( isLeaf(s) // i has a leaf-node sibling
								&& (int)s->ix.size() + (int)q->ix.size() + 1 <= _bucket_size ) ) { // and the combined node would be less than the new bucket size
							++changed;
							svlTreeNode *p = q->parent;
							// copy child indices to parent and change _pts2nodes pointers to parent
							if ( p->left != NULL ) {
								for ( vector<int>::const_iterator j=p->left->ix.begin(); j!=p->left->ix.end(); ++j ) {
									p->ix.push_back(*j);
									_pts2nodes[*j] = p;
								}
								delete p->left;
								p->left = NULL;
							}
							if ( p->right != NULL ) {
								for ( vector<int>::const_iterator j=p->right->ix.begin(); j!=p->right->ix.end(); ++j ) {
									p->ix.push_back(*j);
									_pts2nodes[*j] = p;
								}
								delete p->right;
								p->right = NULL;
							}
						}
					}
				}
			}
		}
	} // else, no changes necessary
}

void svlKdTree::printTree() {
	queue<pair<svlTreeNode*,string> > nodes;
	nodes.push(make_pair(_troot,""));
	cout << _troot->ix.size() << endl;
	int check = (int)nodes.size(), i = 0, dep = 0;
	while ( !nodes.empty() ) {
		svlTreeNode *q = nodes.front().first;
		string str = nodes.front().second;
		nodes.pop();

		cout << "Dep: " << dep << " - " << str << " by dim: " << q->dim << " @ " << q->splitVal << " (" << q->ix.size() << "):\n";//  " << q->box << "\n";
		for ( vector<int>::const_iterator j=q->ix.begin(); j!=q->ix.end(); ++j ) {
			cout << "   :: " << (*_P)[*j] << "\n";// << ( j+1!=q->ix.end() ? "             " : "" );
		}

		// push next nodes onto the stack
		if ( q->left != NULL ) 
			nodes.push(make_pair(q->left,str+"<"));
		if ( q->right != NULL ) 
			nodes.push(make_pair(q->right,str+">"));

		if ( ++i == check ) {
			++dep;
			check += (int)nodes.size();
		}
	}
}

bool svlKdTree::verifyKdTree() {
	return verifyKdTree(_troot); // start search for point at root
}

bool svlKdTree::verifyKdTree(svlTreeNode *node) {
	// Search downtree from current node, verifying that points have been sorted correctly with
	// respect to ancestral *node.

	bool good = true;
	double val = node->splitVal; // ancestral splitting value
	int dim = node->dim;
	queue<svlTreeNode*> nodes; // queues for examining nodes in depth-first search
	if ( node->left != NULL )
		nodes.push(node->left); // check left nodes first; no reason

	while ( good && !nodes.empty() ) { // LEFT!
		svlTreeNode* q = nodes.front();
		nodes.pop();

		if ( (int)q->ix.size() == 1 ) { // branch node
			if ( (*_P)[q->ix[0]][dim] > val ) {
				good = false;
				break;
			}
			if ( q->left != NULL )
				nodes.push(q->left);
			if ( q->right != NULL )
				nodes.push(q->right);
		} else { // bucket node
			for ( vector<int>::const_iterator i=q->ix.begin(); i!=q->ix.end(); ++i ) {
				if ( (*_P)[*i][dim] > val ) {
					good = false;
					break;
				}
			}
		}
	}

	if ( !good ) return good;
	
	// else, check the right nodes
	if ( node->right != NULL )
		nodes.push(node->right);

	while ( good && !nodes.empty() ) { // RIGHT!
		svlTreeNode* q = nodes.front();
		nodes.pop();

		if ( (int)q->ix.size() == 1 ) { // branch node
			if ( (*_P)[q->ix[0]][dim] < val ) {
				good = false;
				break;
			}
			if ( q->left != NULL )
				nodes.push(q->left);
			if ( q->right != NULL )
				nodes.push(q->right);
		} else { // bucket node
			for ( vector<int>::const_iterator i=q->ix.begin(); i!=q->ix.end(); ++i ) {
				if ( (*_P)[*i][dim] < val ) {
					good = false;
					break;
				}
			}
		}
	}

	// Check sorting under the sub nodes
	good = good && ( node->left == NULL ? true : verifyKdTree(node->left) )
				&& ( node->right == NULL ? true : verifyKdTree(node->right) );
	return good;
}

int svlKdTree::findExactNode(int &ix, svlPoint3d &pt, svlTreeNode* &node, int dep) {
	// Find the exact entry of point pt (*_P)[ix] in the tree; return pointer and depth found at.
	if ( node->ix.size() == 1 ) { // if the vector is one element, be quick about it
		if ( node->ix[0] == ix )
			return dep;
		// ix is not at this node; check its children:
		int dim = dep%3; // dimension to compare at
		if ( node->right != NULL && pt.getD(dim) >= (*_P)[node->ix[0]].getD(dim) ) { // search right nodes
			node = node->right;
			int ix1 = findExactNode(ix,pt,node,dep+1);
			if ( ix1 >= 0 )
				return ix1; // return the depth it was found at
			node = node->parent; // else, got nothing: go back up to parent to check left child
		}
		if ( node->left != NULL && pt.getD(dim) <= (*_P)[node->ix[0]].getD(dim) ) { // search left nodes
			node = node->left;
			return findExactNode(ix,pt,node,dep+1); // if -1 here, still return it
		}
	} else { // check the whole vector
		bool here = false;
		for ( vector<int>::const_iterator i=node->ix.begin(); !here && i!=node->ix.end(); ++i ) {
			here = *i == ix;
		}
		if ( here )
			return dep;
	}
	return -1;
}

int svlKdTree::getDepthOfNode(svlTreeNode* &q) {
	if ( q == _troot ) return 0;
	else return 1+getDepthOfNode(q->parent);
}

void svlKdTree::testFindExactAll() {
	bool good = true;
	int i = 0;
	svlTreeNode *q;
	for ( ; good && i<n_pts; ++i ) {
		q = _troot;
		int dep = findExactNode(i,(*_P)[i],q,0);
		//cout << dep << endl;
		good = q == _pts2nodes[i];
	}
	cout << ( good ? "_pts2nodes passes" : "_pts2nodes ***FAILED***" ) << endl;
	if ( !good ) {
		/*cout << i << ": " << q << ", " << _pts2nodes[i] << endl;
		cout << (*_P)[i] << endl;
		cout << "Checking _pts2nodes:\n";
		for ( vector<int>::const_iterator j=_pts2nodes[i]->ix.begin(); j!=_pts2nodes[i]->ix.end(); ++j ) {
			cout << "  " << (*_P)[*j] << "\n";
		}
		cout << "Checking q:\n";
		for ( vector<int>::const_iterator j=q->ix.begin(); j!=q->ix.end(); ++j ) {
			cout << "  " << (*_P)[*j] << "\n";
		}*/
		return;
	}

	clock_t stc = clock();
	double hister = 100;
	for ( int j=0; j<hister; ++j ) {
		for ( int i=0; i<n_pts; ++i ) {
			svlTreeNode *q = _troot;
			int dep = findExactNode(i,(*_P)[i],q,0);
		}
	}
	cout << "findExactNode (per " << hister << "): " << ((double)clock()-stc)/hister/CLOCKS_PER_SEC << endl;
}

void svlKdTree::findAllKNN(int k, double max_dist, vector<vector<int> > &knn) {
	knn = vector<vector<int> >(n_pts,vector<int>()); // reserve index vectors
	for ( int i=0; i<n_pts; ++i ) {
		findKNN(i,k,max_dist,knn[i]);
	}
}

void svlKdTree::findKNN(int ix, int k, double max_dist, vector<int> &ixs) {
	OrdPQ pq;
	if ( (int)ixs.size() != 0 )
		ixs.clear();
	findKNN(ix,k,max_dist,pq);
	if ( pq.empty() )
		return;
	ixs.reserve(pq.size());
	double mmax = pq.top().first; // max distance
	while ( !pq.empty() ) {
		ixs.push_back(pq.top().second);
		// adjust max_dist for nearest-neighbor-related point
		double d = mmax + pq.top().first;
		if ( _max_dists[pq.top().second] < 0.0 || d < _max_dists[pq.top().second] ) {
			_max_dists[pq.top().second] = d;
		}
		pq.pop();
	}
#if MY_DEBUG
	if ( debug ) ++n_checks;
#endif
}

void svlKdTree::findKNN(int ix, int k, double max_dist, vector<pair<double,int> > &dist_ixs) {
	OrdPQ pq;
	findKNN(ix,k,max_dist,pq);
	if ( (int)dist_ixs.size() != 0 )
		dist_ixs.clear();
	dist_ixs.reserve(k);
	double mmax = pq.top().first; // max distance
	while ( !pq.empty() ) {
		dist_ixs.push_back(pq.top());
		// adjust max_dist for nearest-neighbor-related point
		double d = mmax + pq.top().first;
		if ( _max_dists[pq.top().second] < 0.0 || d < _max_dists[pq.top().second] ) {
			_max_dists[pq.top().second] = d;
		}
		pq.pop();
	}
#if MY_DEBUG
	if ( debug ) ++n_checks;
#endif
}

inline void svlKdTree::point2PQ(svlPoint3d &pt, int &k, OrdPQ &pq, double &max_dist, int index) {
	// BTW: doing it this way is actually slower:
	/*bool fig = max_dist < 0.0 || (pt-(*_P)[q->ix[0]]).abs().pt_max() <= max_dist;
	double dist = fig ? sqrt( ( pt - (*_P)[q->ix[0]] ).norm2() ) : 0.0;
	if ( fig && ( max_dist < 0.0 || dist <= max_dist ) ) { ...*/

	double dist = sqrt( ( pt - (*_P)[index] ).norm2() );
	if ( max_dist < 0.0 || dist <= max_dist ) {
		if ( (int)pq.size() < k ) { // queue not full; add in nearest neighbor
			pq.push(make_pair(dist,index));
			if ( (int)pq.size() == k ) // queue now full ...
				max_dist = pq.top().first; // ... can lower max_dist to constrain searches even further
		} else {
			pq.pop(); // remove farthest element
			pq.push(make_pair(dist,index)); // add new element to nearest neighbors
			max_dist = pq.top().first;
		}
	}
}

void svlKdTree::findKNN(int ix, int k, double max_dist, OrdPQ &pq) {
	// Find k nearest neigbhors from point _P[ix] (exact query) with maximum distance max_dist.

	// Find point in tree
	svlPoint3d pt = (*_P)[ix];
	svlTreeNode *qnode = _pts2nodes[ix]; // get node pointer from lookup table
	assert(qnode!=NULL); // otherwise we didn't find the point

	//int dep = getDepthOfNode(qnode); // not needed
	if ( max_dist < 0.0 || _max_dists[ix] < max_dist ) {
		max_dist = _max_dists[ix]; // use our cached max_dist estimate
	}
	queue<svlTreeNode*> nodes; // queue for examining nodes in breadth-first search

	if ( qnode->ix.size() > 1 ) { // search over other points in this bucket for near neighbors
		for ( vector<int>::const_iterator i=qnode->ix.begin(); i!=qnode->ix.end(); ++i ) {
			if ( *i != ix ) { // novel point
				double dist = sqrt( ( pt - (*_P)[*i] ).norm2() );
				if ( max_dist < 0.0 || dist <= max_dist ) {
					if ( (int)pq.size() < k ) { // queue not full; add in nearest neighbor
						pq.push(make_pair(dist,*i));
						if ( (int)pq.size() == k ) // queue now full ...
							max_dist = pq.top().first; // ... can lower max_dist to constrain searches even further
					} else {
						pq.pop(); // remove farthest element
						pq.push(make_pair(dist,*i)); // add new element to nearest neighbors
						max_dist = pq.top().first;
					}
				}
			}
		}
	} else { // branch node: perform downtree search from current node
		if ( qnode->left != NULL )
			nodes.push(qnode->left);
		if ( qnode->right != NULL )
			nodes.push(qnode->right); // push on current node's children

		while ( !nodes.empty() ) {
			svlTreeNode* q = nodes.front();
			nodes.pop();
			if ( q != NULL ) {
				// check the point at this node for nearest-neighbor status
				if ( (int)q->ix.size() == 1 ) { // potential branch point
					double dist = sqrt( ( pt - (*_P)[q->ix[0]] ).norm2() );
					if ( max_dist < 0.0 || dist <= max_dist ) {
						if ( (int)pq.size() < k ) { // queue not full; add in nearest neighbor
							pq.push(make_pair(dist,q->ix[0]));
							if ( (int)pq.size() == k ) // queue now full ...
								max_dist = pq.top().first; // ... can lower max_dist to constrain searches even further
						} else {
							pq.pop(); // remove farthest element
							pq.push(make_pair(dist,q->ix[0])); // add new element to nearest neighbors
							max_dist = pq.top().first;
						}
					}

					if ( q->left != NULL // there are points to check down left fork
						&& ( max_dist < 0.0 || ( pt.getD(q->dim) - q->splitVal ) <= max_dist ) ) { // and they could be within the max distance
						nodes.push(q->left); // add this node to search next time around
					}
					if ( q->right != NULL // there are points to check down right fork
						&& ( max_dist < 0.0 || ( q->splitVal - pt.getD(q->dim) ) <= max_dist ) ) { // and they could be within the max distance
						nodes.push(q->right);
					}
				} else { // bucket node: check entire bucket
					for ( vector<int>::const_iterator i=q->ix.begin(); i!=q->ix.end(); ++i ) {
						double dist = sqrt( ( pt - (*_P)[*i] ).norm2() );
						if ( max_dist < 0.0 || dist <= max_dist ) {
							if ( (int)pq.size() < k ) { // queue not full; add in nearest neighbor
								pq.push(make_pair(dist,*i));
								if ( (int)pq.size() == k ) // queue now full ...
									max_dist = pq.top().first; // ... can lower max_dist to constrain searches even further
							} else {
								pq.pop(); // remove farthest element
								pq.push(make_pair(dist,*i)); // add new element to nearest neighbors
								max_dist = pq.top().first;
							}
						}
					}
				}
			}
		}
	}

	if ( max_dist > 0.0 && qnode->box.mDFPWOB(pt,_troot->box) >= max_dist )
		return;
	
	// Perform uptree search
	while ( true ) {
		svlTreeNode *par = qnode->parent;
		if ( par == NULL )
			break; // found node was troot already
		
		// check if parent and its other branch could have points within max_dist (the hyperplane rule)
		if ( max_dist < 0.0 || abs(par->splitVal - pt[par->dim]) <= max_dist ) {
			double dist = sqrt( ( pt - (*_P)[par->ix[0]] ).norm2() );
			if ( max_dist < 0.0 || dist <= max_dist ) {
				if ( (int)pq.size() < k ) { // queue not full; add in nearest neighbor
					pq.push(make_pair(dist,par->ix[0]));
					if ( (int)pq.size() == k ) // queue now full ...
						max_dist = pq.top().first; // ... can lower max_dist to constrain searches even further
				} else {
					pq.pop(); // remove farthest element
					pq.push(make_pair(dist,par->ix[0])); // add new element to nearest neighbors
					max_dist = pq.top().first;
				}
			}
			//assert((int)par->ix.size()==1); // no parent can be a bucket

			// perform downtree search on other branch
			svlTreeNode *other = qnode == par->left ? par->right : par->left;
			if ( other != NULL ) {
				nodes.push(other); // search the other branch besides qnode
				while ( !nodes.empty() ) {
					svlTreeNode *q = nodes.front();
					nodes.pop();
					
					if ( q != NULL ) {
						// check the point at this node for nearest-neighbor status
						if ( (int)q->ix.size() == 1 ) { // potential branch point
							double dist = sqrt( ( pt - (*_P)[q->ix[0]] ).norm2() );
							if ( max_dist < 0.0 || dist <= max_dist ) {
								if ( (int)pq.size() < k ) { // queue not full; add in nearest neighbor
									pq.push(make_pair(dist,q->ix[0]));
									if ( (int)pq.size() == k ) // queue now full ...
										max_dist = pq.top().first; // ... can lower max_dist to constrain searches even further
								} else {
									pq.pop(); // remove farthest element
									pq.push(make_pair(dist,q->ix[0])); // add new element to nearest neighbors
									max_dist = pq.top().first;
								}
							}
							if ( q->left != NULL // there are points to check down left fork
								&& ( max_dist < 0.0 || ( pt.getD(q->dim) - q->splitVal ) <= max_dist ) ) { // and they could be within the max distance
								nodes.push(q->left); // add this node to search next time around
							}
							if ( q->right != NULL // there are points to check down right fork
								&& ( max_dist < 0.0 || ( q->splitVal - pt.getD(q->dim) ) <= max_dist ) ) { // and they could be within the max distance
								nodes.push(q->right);
							}
						} else { // bucket node: check entire bucket
							for ( vector<int>::const_iterator i=q->ix.begin(); i!=q->ix.end(); ++i ) {
								double dist = sqrt( ( pt - (*_P)[*i] ).norm2() );
								if ( max_dist < 0.0 || dist <= max_dist ) {
									if ( (int)pq.size() < k ) { // queue not full; add in nearest neighbor
										pq.push(make_pair(dist,*i));
										if ( (int)pq.size() == k ) // queue now full ...
											max_dist = pq.top().first; // ... can lower max_dist to constrain searches even further
									} else {
										pq.pop(); // remove farthest element
										pq.push(make_pair(dist,*i)); // add new element to nearest neighbors
										max_dist = pq.top().first;
									}
								}
							}
						}
					}
				}
			}
		}
		if ( par == _troot // we just checked the root, or all points
			|| ( max_dist > 0.0 && par->box.mDFPWOB(pt,_troot->box) >= max_dist ) ) // bounding box is greater than max_dist
			return;
		qnode = par; // move one level up the tree
	}
}

void svlKdTree::findKNNd(int ix, int k, double max_dist, OrdPQ &pq) {
	// Find k nearest neigbhors from point _P[ix] (exact query) with maximum distance max_dist.

	// Find point in tree
	svlPoint3d pt = (*_P)[ix];
	svlTreeNode *qnode = _pts2nodes[ix]; // get node pointer from lookup table
	assert(qnode!=NULL); // otherwise we didn't find the point
	int dep = getDepthOfNode(qnode);
#if MY_DEBUG_vx2
	cout << ":> " << max_dist << " -> ";
#endif
	if ( max_dist < 0.0 || _max_dists[ix] < max_dist )
		max_dist = _max_dists[ix]; // use our cached max_dist estimate
#if MY_DEBUG_vx2
	cout << max_dist << "\n";
#endif
	//cout << " (" << max_dist << ")";

	queue<svlTreeNode*> nodes; // queue for examining nodes in breadth-first search

	// Search over other points in the bucket for near neighbors
	if ( qnode->ix.size() > 1 ) {
		for ( vector<int>::const_iterator i=qnode->ix.begin(); i!=qnode->ix.end(); ++i ) {
			if ( *i != ix ) { // novel point
#define ACH 0
#if ACH
				point2PQ(pt,k,pq,max_dist,*i);
#else
				double dist = sqrt( ( pt - (*_P)[*i] ).norm2() );
				if ( max_dist < 0.0 || dist <= max_dist ) {
					if ( (int)pq.size() < k ) { // queue not full; add in nearest neighbor
						pq.push(make_pair(dist,*i));
						if ( (int)pq.size() == k ) // queue now full ...
							max_dist = pq.top().first; // ... can lower max_dist to constrain searches even further
					} else {
						pq.pop(); // remove farthest element
						pq.push(make_pair(dist,*i)); // add new element to nearest neighbors
						max_dist = pq.top().first;
					}
				}
#endif
			}
		}
#if MY_DEBUG
		if ( debug ) checks += (long)qnode->ix.size()-1;
#endif
	} else { // branch node: perform downtree search from current node
		//cout << qnode->box << "\n ->> " << qnode->box.mDFPWOB(pt,_troot->box) << endl;
	
		if ( qnode->left != NULL )
			nodes.push(qnode->left);
		if ( qnode->right != NULL )
			nodes.push(qnode->right); // push on current node's children
		nodes.push(NULL); // next-depth transition pointer

		int t_dep = dep+1;
		while ( !nodes.empty() ) {
#if MY_DEBUG
			if ( debug ) checks += (long)nodes.size();
#endif
			svlTreeNode* q = nodes.front();
			nodes.pop();
			
			if ( q != NULL ) {
				// check the point at this node for nearest-neighbor status
				if ( (int)q->ix.size() == 1 ) { // potential branch point
#if ACH
					point2PQ(pt,k,pq,max_dist,q->ix[0]);
#else
					double dist = sqrt( ( pt - (*_P)[q->ix[0]] ).norm2() );
					if ( max_dist < 0.0 || dist <= max_dist ) {
						if ( (int)pq.size() < k ) { // queue not full; add in nearest neighbor
							pq.push(make_pair(dist,q->ix[0]));
							if ( (int)pq.size() == k ) // queue now full ...
								max_dist = pq.top().first; // ... can lower max_dist to constrain searches even further
						} else {
							pq.pop(); // remove farthest element
							pq.push(make_pair(dist,q->ix[0])); // add new element to nearest neighbors
							max_dist = pq.top().first;
						}
					}
#endif

					int dim = t_dep%3; // dimension to compare at
					if ( q->left != NULL // there are points to check down left fork
						&& ( max_dist < 0.0 || ( pt.getD(dim) - q->splitVal ) <= max_dist ) ) { // and they could be within the max distance
						nodes.push(q->left); // add this node to search next time around
					}
					if ( q->right != NULL // there are points to check down right fork
						&& ( max_dist < 0.0 || ( q->splitVal - pt.getD(dim) ) <= max_dist ) ) { // and they could be within the max distance
						nodes.push(q->right);
					}
				} else { // bucket node: check entire bucket
					for ( vector<int>::const_iterator i=q->ix.begin(); i!=q->ix.end(); ++i ) {
#if ACH
						point2PQ(pt,k,pq,max_dist,*i);
#else
						double dist = sqrt( ( pt - (*_P)[*i] ).norm2() );
						if ( max_dist < 0.0 || dist <= max_dist ) {
							if ( (int)pq.size() < k ) { // queue not full; add in nearest neighbor
								pq.push(make_pair(dist,*i));
								if ( (int)pq.size() == k ) // queue now full ...
									max_dist = pq.top().first; // ... can lower max_dist to constrain searches even further
							} else {
								pq.pop(); // remove farthest element
								pq.push(make_pair(dist,*i)); // add new element to nearest neighbors
								max_dist = pq.top().first;
							}
						}
#endif
					}
				}
			} else if ( !nodes.empty() ) { // there is another depth level to examine
				nodes.push(NULL);
				++t_dep;
			}
		}
	}

#if MY_DEBUG_vx2
	if ( (int)pq.size() > 0 ) {
		cout << "{\n   " << pt
			 << "\n   d: " << dep << ", s: " << pq.size() << ", top: " << pq.top().first
			 << "\n   " << qnode->box << "\n   " << qnode->box.mDFPWOB(pt,_troot->box) << " vs " << max_dist
			 << "\n}\n";
	}
#endif
	if ( max_dist > 0.0 && qnode->box.mDFPWOB(pt,_troot->box) >= max_dist ) {
		return;
	}
	// Perform uptree search
	while ( true ) {
		svlTreeNode *par = qnode->parent;
		if ( par == NULL )
			break; // found node was _troot already
		--dep;
#if MY_DEBUG
	if ( debug ) {
		++checks;
		++n_levels;
	}
#endif
		// check if parent and its other branch could have points within max_dist (the hyperplane rule)
		int dim = dep%3;
		if ( max_dist < 0.0 || abs(par->splitVal - pt[dim]) <= max_dist ) {
#if ACH
			point2PQ(pt,k,pq,max_dist,par->ix[0]); // compare parent point manually to q
#else
			double dist = sqrt( ( pt - (*_P)[par->ix[0]] ).norm2() );
			if ( max_dist < 0.0 || dist <= max_dist ) {
				if ( (int)pq.size() < k ) { // queue not full; add in nearest neighbor
					pq.push(make_pair(dist,par->ix[0]));
					if ( (int)pq.size() == k ) // queue now full ...
						max_dist = pq.top().first; // ... can lower max_dist to constrain searches even further
				} else {
					pq.pop(); // remove farthest element
					pq.push(make_pair(dist,par->ix[0])); // add new element to nearest neighbors
					max_dist = pq.top().first;
				}
			}
#endif

			//assert((int)par->ix.size()==1); // if par->ix.size() > 1, impossible

			// perform downtree search on other branch
			svlTreeNode *other = qnode == par->left ? par->right : par->left;
			if ( other != NULL ) {
				nodes.push(other); // search the other branch besides qnode
				nodes.push(NULL); // next depth transition point
				int t_dep = dep+1; // temporary depth for downtree search
				while ( !nodes.empty() ) {
#if MY_DEBUG
	if ( debug ) checks += (long)nodes.size();
#endif
					svlTreeNode *q = nodes.front();
					nodes.pop();
					
					if ( q != NULL ) {
						// check the point at this node for nearest-neighbor status
						if ( (int)q->ix.size() == 1 ) { // potential branch point
#if ACH
							point2PQ(pt,k,pq,max_dist,q->ix[0]); // compare parent point manually to q
#else
							double dist = sqrt( ( pt - (*_P)[q->ix[0]] ).norm2() );
							if ( max_dist < 0.0 || dist <= max_dist ) {
								if ( (int)pq.size() < k ) { // queue not full; add in nearest neighbor
									pq.push(make_pair(dist,q->ix[0]));
									if ( (int)pq.size() == k ) // queue now full ...
										max_dist = pq.top().first; // ... can lower max_dist to constrain searches even further
								} else {
									pq.pop(); // remove farthest element
									pq.push(make_pair(dist,q->ix[0])); // add new element to nearest neighbors
									max_dist = pq.top().first;
								}
							}
#endif
							int dim = t_dep%3; // dimension to compare at
							if ( q->left != NULL // there are points to check down left fork
								&& ( max_dist < 0.0 || ( pt.getD(dim) - q->splitVal ) <= max_dist ) ) { // and they could be within the max distance
								nodes.push(q->left); // add this node to search next time around
							}
							if ( q->right != NULL // there are points to check down right fork
								&& ( max_dist < 0.0 || ( q->splitVal - pt.getD(dim) ) <= max_dist ) ) { // and they could be within the max distance
								nodes.push(q->right);
							}
						} else { // bucket node: check entire bucket
							for ( vector<int>::const_iterator i=q->ix.begin(); i!=q->ix.end(); ++i ) {
#if ACH
								point2PQ(pt,k,pq,max_dist,*i);
#else
								double dist = sqrt( ( pt - (*_P)[*i] ).norm2() );
								if ( max_dist < 0.0 || dist <= max_dist ) {
									if ( (int)pq.size() < k ) { // queue not full; add in nearest neighbor
										pq.push(make_pair(dist,*i));
										if ( (int)pq.size() == k ) // queue now full ...
											max_dist = pq.top().first; // ... can lower max_dist to constrain searches even further
									} else {
										pq.pop(); // remove farthest element
										pq.push(make_pair(dist,*i)); // add new element to nearest neighbors
										max_dist = pq.top().first;
									}
								}
#endif
							}
						}
					} else if ( !nodes.empty() ) { // there is another depth level to examine
						nodes.push(NULL);
						++t_dep;
					} // else, let the thread expire
				}
			}
		}
#if MY_DEBUG_vx2
	cout << "{\n   " << pt
		 << "\n   d: " << dep << ", s: " << pq.size() << ", top: " << pq.top().first
		 << "\n   " << qnode->box << "\n   " << qnode->box.mDFPWOB(pt,_troot->box) << " vs " << max_dist
		 << "\n}\n";
#endif
		if ( par == _troot // we just checked the root, or all points
			|| ( max_dist > 0.0 && par->box.mDFPWOB(pt,_troot->box) >= max_dist ) ) { // bounding box is greater than max_dist
			return;
		}
		qnode = par; // move one level up the tree
	}
}

void svlKdTree::findKNN(svlPoint3d &pt, int k, double max_dist, vector<int> &ixs) {
	OrdPQ pq;
	findKNN(pt,k,max_dist,pq);
	if ( (int)ixs.size() != 0 )
		ixs.clear();
	ixs.reserve(k);
	while ( !pq.empty() ) {
		ixs.push_back(pq.top().second);
		pq.pop();
	}
}

void svlKdTree::findKNN(svlPoint3d &pt, int k, double max_dist, vector<pair<double,int> > &dist_ixs) {
	OrdPQ pq;
	findKNN(pt,k,max_dist,pq);
	if ( (int)dist_ixs.size() != 0 )
		dist_ixs.clear();
	dist_ixs.reserve(k);
	while ( !pq.empty() ) {
		dist_ixs.push_back(pq.top());
		pq.pop();
	}
}

void svlKdTree::findKNN(svlPoint3d &pt, int k, double max_dist, OrdPQ &pq) {
	// Find k nearest neigbhors from point pt with maximum distance max_dist.
	int dep = 0;
	queue<svlTreeNode*> nodes; // queues for examining nodes in depth-first search
	nodes.push(_troot); // start search for point at root

	//cout << "Searching for point: " << pt << endl;
	// downtree search from here
	while ( !nodes.empty() ) {
		svlTreeNode* q = nodes.front();
		nodes.pop();
		
		//cout << " Branch: " << q->ix[0] << ": " << (*_P)[q->ix[0]] << endl;
		if ( (int)q->ix.size() == 1 ) { // potential branch point
			point2PQ(pt,k,pq,max_dist,q->ix[0]);
			if ( q->left != NULL // there are points to check down left fork
				 && ( max_dist < 0.0 || ( pt[q->dim] - q->splitVal ) <= max_dist ) ) // and they could be within the max distance
				nodes.push(q->left); // add this node to search next time around
			if ( q->right != NULL // there are points to check down right fork
				 && ( max_dist < 0.0 || ( q->splitVal - pt[q->dim] ) <= max_dist ) ) // and they could be within the max distance
				nodes.push(q->right);
		} else // bucket node: check entire bucket
			for ( vector<int>::const_iterator i=q->ix.begin(); i!=q->ix.end(); ++i )
				point2PQ(pt,k,pq,max_dist,*i);
	}
}

void svlKdTree::start_debug() {
	checks = 0;
	n_checks = 0;
	n_levels = 0;
	debug = true;
}
void svlKdTree::stop_debug() {
	debug = false;
}
void svlKdTree::print_debug() {
	cout << n_checks << " checks, " << (double)(checks)/(double)(n_checks) << " comps/; " << (double)(n_levels)/(double)(n_checks) << " levs/\n";
}
void svlKdTree::set_bucket_size(int bucket_size) {
	this->~svlKdTree(); // destroy whatever exists
	_bucket_size = bucket_size;
}
int svlKdTree::get_bucket_size() {
	return _bucket_size;
}

bool containsIx(svlTreeNode* q, int ix) {
	if ( q->ix.size() == 1 ) { // branch
		if ( q->ix[0] == ix ) return true;
		if ( q->left != NULL )
			if ( containsIx(q->left,ix) )
				return true;
		//cout << q->ix[0] << " left no dups " << ix << "\n";
		if ( q->right != NULL )
			if ( containsIx(q->right,ix) )
				return true;
		//cout << q->ix[0] << " right no dups " << ix << "\n";
		return false;
	} else {
		bool contains = false;
		for ( vector<int>::const_iterator i=q->ix.begin(); !contains && i!=q->ix.end(); ++i ) {
			contains = *i == ix;
		}
		//cout << q << " bucket no dups " << ix << "\n";
		return contains;
	}
}

bool svlKdTree::searchDuplicateIxs() {
	// search for any duplicate indices in the tree
	// caveat: only works going downtree
	queue<svlTreeNode*> nodes;
	nodes.push(_troot);
	while ( !nodes.empty() ) {
		svlTreeNode* q = nodes.front();
		nodes.pop();

		if ( q->ix.size() == 1 ) { // branch
			if ( q->left != NULL ) {
				if ( containsIx(q->left,q->ix[0]) ) {
					cout << "Node " << q->ix[0] << " had duplicates.\n";
					return true;
				}
				nodes.push(q->left);
			}
			if ( q->right!= NULL ) {
				if ( containsIx(q->right,q->ix[0]) ) {
					cout << "Node " << q->ix[0] << " had duplicates.\n";
					return true;
				}
				nodes.push(q->right);
			}
		} else { // bucket; check for duplicates within bucket
			bool duplicates = false;
			for ( vector<int>::const_iterator i=q->ix.begin(); !duplicates && i!=q->ix.end(); ++i ) {
				for ( vector<int>::const_iterator j=i+1; !duplicates && j!=q->ix.end(); ++j ) {
					duplicates = *i == *j;
				}
			}
			if ( duplicates ) {
				cout << "Bucket node had duplicates\n";
				return true;
			}
		}
	}
	return false;
}

bool svlKdTree::isLeaf(svlTreeNode* q) {
	return q->left == NULL && q->right == NULL;
}

int svlKdTree::get_n_pts() {
	return n_pts;
}

svlTreeNode* svlKdTree::getSibling(svlTreeNode *q) {
	if ( q->parent == NULL ) return NULL;
	return ( q->parent->left == q ? q->parent->right : q->parent->left );
}

