/******************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlPointNd.h
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Ian Goodfellow <ia3n@stanford.edu>
**				Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**  Basic N-d point class. Supports simple mathematical operations. The y-axis
**  is defined as up, so panning is about the y-axis, tilting is about the
**  x-axis and rolling is about the z-axis.
** 
*****************************************************************************/

#pragma once

#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
#define _CRT_SECURE_NO_DEPRECATE
#include <limits>
#undef max
#undef min
#endif

#include <cassert>
#include <cmath>
#include <iostream>
#include <limits>
using namespace std;

template<class V = double, int N = 3>
class svlPointNd {
public:
	V _vals[N];

public:
	inline svlPointNd() {
		//_vals = new V[N];
		for ( int i=0; i<N; ++i ) _vals[i] = 0;
	}
	inline svlPointNd(V v) {
		for ( int i=0; i<N; ++i ) _vals[i] = v;
	}
	inline svlPointNd(V vals[]) {
		for ( int i=0; i<N; ++i ) _vals[i] = vals[i];
	}
	template<int N1> inline svlPointNd(const svlPointNd<V,N1>& p) {
		int i=0;
		for ( ; i<N && i<N1; ++i ) _vals[i] = p._vals[i];
		for ( ; i<N; ++i ) _vals[i] = 0;
	}
	inline svlPointNd(V v0, V v1);
	inline svlPointNd(V v0, V v1, V v2);
	virtual ~svlPointNd() {}

	inline int size() const;
	inline V getD(int d) const;
	inline void setD(int d, V val);
	inline void set(V val);
	inline void set(V vals[]);
	inline void set(V v0, V v1);
	inline void set(V v0, V v1, V v2);
	inline V& operator[](int i);
	inline V operator[](int i) const;
	inline V& x();
	inline V x() const;
	inline V& y();
	inline V y() const;
	inline V& z();
	inline V z() const;
	inline V& t();
	inline V t() const;

	inline V norm() const;
	inline V norm2() const;
	inline void set_min_pt();
	inline void set_max_pt();
	template<class V1, int N1> inline V dot(const svlPointNd<V1,N1>& p) const;
	inline svlPointNd<V,N> cross(const svlPointNd<V,N> & p) const;
	inline svlPointNd<V,N>& normalize();
	inline svlPointNd<V,N>& ptwise_sqrt();
	inline svlPointNd<V,N>& abs();
	inline V pt_min() const;
	inline V pt_max() const;
	template<int N1, int N2> friend inline svlPointNd<V,N2> min2(const svlPointNd<V,N>& p, const svlPointNd<V,N1>& q);
	template<int N1, int N2> friend inline svlPointNd<V,N2> max2(const svlPointNd<V,N>& p, const svlPointNd<V,N1>& q);

	template<class V1, int N1>	inline svlPointNd<V,N>& min_with(const svlPointNd<V1,N1>& p);
	template<class V1>			inline svlPointNd<V,N>& min_with(V1 val);
	template<class V1, int N1>	inline svlPointNd<V,N>& max_with(const svlPointNd<V1,N1>& p);
	template<class V1>			inline svlPointNd<V,N>& max_with(V1 val);

	inline svlPointNd<V,N>& pan(V theta);
	inline svlPointNd<V,N>& pan(V cos_theta, V sin_theta);
	inline svlPointNd<V,N>& pan(V theta, svlPointNd<V,N>& c);
	inline svlPointNd<V,N>& tilt(V theta);
	inline svlPointNd<V,N>& tilt(V cos_theta, V sin_theta);
	inline svlPointNd<V,N>& tilt(V theta, svlPointNd<V,N>& c);
	inline svlPointNd<V,N>& roll(V theta);
	inline svlPointNd<V,N>& roll(V theta, svlPointNd<V,N>& c);

	template<int N1> inline svlPointNd<V,N>& operator=(const svlPointNd<V,N1>& p);
	inline svlPointNd<V,N>& operator=(const V d);
	inline bool operator>(const V d);
	inline bool operator>=(const V d);
	inline bool operator<(const V d);
	inline bool operator<=(const V d);
	template<int N1> inline svlPointNd<V,N>& operator+=(const svlPointNd<V,N1>& p);
	inline svlPointNd<V,N>& operator+=(const V d);
	template<int N1> inline svlPointNd<V,N>& operator-=(const svlPointNd<V,N1>& p);
	inline svlPointNd<V,N>& operator-=(const V d);
	template<int N1> inline svlPointNd<V,N>& operator*=(const svlPointNd<V,N1>& p);
	inline svlPointNd<V,N>& operator*=(const V d);
	template<int N1> inline svlPointNd<V,N>& operator/=(const svlPointNd<V,N1>& p);
	inline svlPointNd<V,N>& operator/=(const V d);
	template<int N1> inline bool operator==(const svlPointNd<V,N1>& p) const;
	inline bool operator==(const V d) const;
	template<int N1> inline bool operator!=(const svlPointNd<V,N1>& p) const;
	inline bool operator!=(const V d) const;

	template<class V1>			friend inline svlPointNd<V,N> operator+(const svlPointNd<V,N>& p, V1 d);
	template<class V1>			friend inline svlPointNd<V,N> operator+(V1 d, const svlPointNd<V,N>& p);
    template<class V1>			friend inline svlPointNd<V,N> operator-(const svlPointNd<V,N>& p);
    template<class V1>			friend inline svlPointNd<V,N> operator+(const svlPointNd<V,N>& p, V1 d);	
    template<class V1>			friend inline svlPointNd<V,N> operator+(V1 d, const svlPointNd<V,N>& p);
	template<class V1, int N1>	friend inline svlPointNd<V,N> operator+(const svlPointNd<V,N>& p, const svlPointNd<V1,N1>& q);
	template<class V1>			friend inline svlPointNd<V,N> operator-(const svlPointNd<V,N>& p, V1 d);
	template<class V1>			friend inline svlPointNd<V,N> operator-(V1 d, const svlPointNd<V,N>& p);
	template<class V1, int N1>	friend inline svlPointNd<V,N> operator-(const svlPointNd<V,N>& p, const svlPointNd<V1,N1>& q);
	template<class V1>			friend inline svlPointNd<V,N> operator*(const svlPointNd<V,N>& p, V1 d);
	template<class V1>			friend inline svlPointNd<V,N> operator*(V1 d, const svlPointNd<V,N>& p);
	template<class V1, int N1>	friend inline svlPointNd<V,N> operator*(const svlPointNd<V,N>& p, const svlPointNd<V1,N1>& q);
	template<class V1>			friend inline svlPointNd<V,N> operator/(const svlPointNd<V,N>& p, V1 d);
	template<class V1>			friend inline svlPointNd<V,N> operator/(V1 d, const svlPointNd<V,N>& p);
	template<class V1, int N1>	friend inline svlPointNd<V,N> operator/(const svlPointNd<V,N>& p, const svlPointNd<V1,N1>& q);

	friend std::ostream& operator<<(std::ostream& os, const svlPointNd<V,N>& p) {
		os << p._vals[0];
		for ( int i=1; i<N; ++i ) os << " " << p._vals[i];
		return os;
	}
	friend std::istream& operator>>(std::istream& is, svlPointNd<V,N>& p) {
		for ( int i=0; i<N; ++i ) is >> p._vals[i];
		return is;
	}
};

/* Implementation ***********************************************************/

template<class V, int N>
inline svlPointNd<V,N>::svlPointNd(V v0, V v1) {
	_vals[0] = v0;
	_vals[1] = v1;
}

template<class V, int N>
inline svlPointNd<V,N>::svlPointNd(V v0, V v1, V v2) {
	_vals[0] = v0;
	_vals[1] = v1;
	_vals[2] = v2;
}

// ****** Accessors ******

template<class V, int N>
inline int svlPointNd<V,N>::size() const {
	return N;
}

template<class V, int N>
inline V svlPointNd<V,N>::getD(int d)  const {
	//assert(d<N);
	return _vals[d];
}

template<class V, int N>
inline void svlPointNd<V,N>::setD(int d, V val) {
	_vals[d] = val;
}

template<class V, int N>
inline void svlPointNd<V,N>::set(V val) {
	for ( int i=0; i<N; ++i ) _vals[i] = val;
}

template<class V, int N>
inline void svlPointNd<V,N>::set(V vals[]) {
	for ( int i=0; i<N; ++i ) _vals[i] = vals[i];
}

template<class V, int N>
inline void svlPointNd<V,N>::set(V v0, V v1) {
	_vals[0] = v0;
	_vals[1] = v1;
}

template<class V, int N>
inline void svlPointNd<V,N>::set(V v0, V v1, V v2) {
	_vals[0] = v0;
	_vals[1] = v1;
	_vals[2] = v2;
}

template<class V, int N>
inline V& svlPointNd<V,N>::operator[](int i) {
	//assert(i<N);
	return this->_vals[i];
}

template<class V, int N>
inline V svlPointNd<V,N>::operator[](int i) const {
	//assert(i<N);
	return this->_vals[i];
}

template<class V, int N>
inline V& svlPointNd<V,N>::x() {
	return this->_vals[0];
}

template<class V, int N>
inline V svlPointNd<V,N>::x() const {
	return this->_vals[0];
}

template<class V, int N>
inline V& svlPointNd<V,N>::y() {
	return this->_vals[1];
}

template<class V, int N>
inline V svlPointNd<V,N>::y() const {
	return this->_vals[1];
}

template<class V, int N>
inline V& svlPointNd<V,N>::z() {
	return this->_vals[2];
}

template<class V, int N>
inline V svlPointNd<V,N>::z() const {
	return this->_vals[2];
}

template<class V, int N>
inline V& svlPointNd<V,N>::t() {
	return this->_vals[3];
}

template<class V, int N>
inline V svlPointNd<V,N>::t() const {
	return this->_vals[3];
}

template<class V, int N>
inline V svlPointNd<V,N>::norm() const {
	V ret = 0;
	for ( int i=0; i<N; ++i ) ret += _vals[i]*_vals[i];
	return sqrt((double)ret);
}

template<class V, int N>
inline V svlPointNd<V,N>::norm2() const {
	V ret = 0;
	for ( int i=0; i<N; ++i ) ret += _vals[i]*_vals[i];
	return ret;
}

template<class V, int N>
inline void svlPointNd<V,N>::set_min_pt() {
	for ( int i=0; i<N; ++i )
		_vals[i] = numeric_limits<V>::min();
}

template<class V, int N>
inline void svlPointNd<V,N>::set_max_pt() {
	for ( int i=0; i<N; ++i )
		_vals[i] = numeric_limits<V>::max();
}

template<class V, int N>
template<class V1, int N1>
inline V svlPointNd<V,N>::dot(const svlPointNd<V1,N1>& p) const {
	V ret = 0;
	for ( int i=0; i<N && i<N1; ++i ) ret += _vals[i]*p._vals[i];
	return ret;
}

template<class V, int N>
inline svlPointNd<V,N> svlPointNd<V,N>::cross(const svlPointNd<V,N> & p) const	{
	assert(N==3);
	return svlPointNd<V,N>(_vals[1]*p._vals[2] - _vals[2]*p._vals[1],
		_vals[2]*p._vals[0] - _vals[0]*p._vals[2],
		_vals[0]*p._vals[1] - _vals[1]*p._vals[0]);
}

template<class V, int N>
inline svlPointNd<V,N>& svlPointNd<V,N>::normalize() {
	V norm = sqrt(norm2());
	for ( int i=0; i<N; ++i ) _vals[i] /= norm;
	return *this;
}

template<class V, int N>
inline svlPointNd<V,N>& svlPointNd<V,N>::ptwise_sqrt() {
	for ( int i=0; i<N; ++i ) _vals[i] = sqrt(_vals[i]);
	return *this;
}

template<class V, int N>
inline svlPointNd<V,N>& svlPointNd<V,N>::abs() {
	for ( int i=0; i<N; ++i ) _vals[i] = _vals[i] < 0 ? -1*_vals[i] : _vals[i];
	return *this;
}

template<class V, int N>
inline V svlPointNd<V,N>::pt_min() const {
	V mint = _vals[0];
	for ( int i=1; i<N; ++i ) mint = _vals[i] < mint ? _vals[i] : mint;
	return mint;
}

template<class V, int N>
inline V svlPointNd<V,N>::pt_max() const {
	V maxt = _vals[0];
	for ( int i=1; i<N; ++i ) maxt = _vals[i] > maxt ? _vals[i] : maxt;
	return maxt;
}

template<class V, int N, int N1, int N2>
inline svlPointNd<V,N2> min2(const svlPointNd<V,N>& p, const svlPointNd<V,N1>& q) {
	N2 = N < N1 ? N : N1;
	V *vals = new V[N2];
	for ( int i=0; i<N2; ++i ) vals[i] = p[i] < q[i] ? p[i] : q[i];
	svlPointNd<V,N2> ret(vals);
	delete[] vals;
	return ret;
}

template<class V, int N, int N1, int N2>
inline svlPointNd<V,N2> max2(const svlPointNd<V,N>& p, const svlPointNd<V,N1>& q) {
	N2 = N < N1 ? N : N1;
	V *vals = new V[N2];
	for ( int i=0; i<N2; ++i ) vals[i] = p[i] > q[i] ? p[i] : q[i];
	svlPointNd<V,N2> ret(vals);
	delete[] vals;
	return ret;
}

template<class V, int N>
template<class V1, int N1>
inline svlPointNd<V,N>& svlPointNd<V,N>::min_with(const svlPointNd<V1,N1>& p) {
	for ( int i=0; i<N && i<N1; ++i ) _vals[i] = _vals[i] < (V)p._vals[i] ? _vals[i] : (V)p._vals[i];
	return *this;
}

template<class V, int N>
template<class V1>
inline svlPointNd<V,N>& svlPointNd<V,N>::min_with(V1 val) {
	for ( int i=0; i<N; ++i ) _vals[i] = _vals[i] < (V)val ? _vals[i] : (V)val;
	return *this;
}

template<class V, int N>
template<class V1, int N1>
inline svlPointNd<V,N>& svlPointNd<V,N>::max_with(const svlPointNd<V1,N1>& p) {
	for ( int i=0; i<N && i<N1; ++i ) _vals[i] = _vals[i] > (V)p._vals[i] ? _vals[i] : (V)p._vals[i];
	return *this;
}

template<class V, int N>
template<class V1>
inline svlPointNd<V,N>& svlPointNd<V,N>::max_with(V1 val) {
	for ( int i=0; i<N; ++i ) _vals[i] = _vals[i] > (V)val ? _vals[i] : (V)val;
	return *this;
}

template<class V, int N>
inline svlPointNd<V,N>& svlPointNd<V,N>::pan(V theta) {
	assert(N>=3);
	V nx = _vals[0] * cos(theta) - _vals[2] * sin(theta);
	V nz = _vals[0] * sin(theta) + _vals[2] * cos(theta);
	_vals[0] = nx; _vals[2] = nz;
	return *this;
}

template<class V, int N>
inline svlPointNd<V,N>& svlPointNd<V,N>::pan(V cos_theta, V sin_theta) {
	assert(N>=3);
	V nx = _vals[0] * cos_theta - _vals[2] * sin_theta;
	V nz = _vals[0] * sin_theta + _vals[2] * cos_theta;
	_vals[0] = nx; _vals[2] = nz;
	return *this;
}

template<class V, int N>
inline svlPointNd<V,N>& svlPointNd<V,N>::pan(V theta, svlPointNd<V,N>& c) {
	assert(N>=3);
	V nx = (_vals[0] - c._vals[0]) * cos(theta) - (_vals[2] - c._vals[2]) * sin(theta);
	V nz = (_vals[0] - c._vals[0]) * sin(theta) + (_vals[2] - c._vals[2]) * cos(theta);
	_vals[0] = c._vals[0] + nx; _vals[2] = c._vals[2] + nz;    

	return *this;
}

template<class V, int N>
inline svlPointNd<V,N>& svlPointNd<V,N>::tilt(V theta) {
	assert(N>=3);
	V ny = _vals[1] * cos(theta) - _vals[2] * sin(theta);
	V nz = _vals[1] * sin(theta) + _vals[2] * cos(theta);
	_vals[1] = ny; _vals[2] = nz;

	return *this;
}

template<class V, int N>
inline svlPointNd<V,N>& svlPointNd<V,N>::tilt(V cos_theta, V sin_theta) {
	assert(N>=3);
	V ny = _vals[1] * cos_theta - _vals[2] * sin_theta;
	V nz = _vals[1] * sin_theta + _vals[2] * cos_theta;
	_vals[1] = ny; _vals[2] = nz;

	return *this;
}

template<class V, int N>
inline svlPointNd<V,N>& svlPointNd<V,N>::tilt(V theta, svlPointNd<V,N>& c) {
	assert(N>=3);
	V ny = (_vals[1] - c._vals[1]) * cos(theta) - (_vals[2] - c._vals[2]) * sin(theta);
	V nz = (_vals[1] - c._vals[1]) * sin(theta) + (_vals[2] - c._vals[2]) * cos(theta);
	_vals[1] = c._vals[1] + ny; _vals[2] = c._vals[2] + nz;    

	return *this;
}

template<class V, int N>
inline svlPointNd<V,N>& svlPointNd<V,N>::roll(V theta) {
	assert(N>=3);
	V nx = _vals[0] * cos(theta) - _vals[1] * sin(theta);
	V ny = _vals[0] * sin(theta) + _vals[1] * cos(theta);
	_vals[0] = nx; _vals[1] = ny;

	return *this;
}

template<class V, int N>
inline svlPointNd<V,N>& svlPointNd<V,N>::roll(V theta, svlPointNd<V,N>& c) {
	assert(N>=3);
	V nx = (_vals[0] - c._vals[0]) * cos(theta) - (_vals[1] - c._vals[1]) * sin(theta);
	V ny = (_vals[0] - c._vals[0]) * sin(theta) + (_vals[1] - c._vals[1]) * cos(theta);
	_vals[0] = c._vals[0] + nx; _vals[1] = c._vals[1] + ny;    

	return *this;
}

/* Operators ****************************************************************/

template<class V, int N>
template<int N1>
inline svlPointNd<V,N>& svlPointNd<V,N>::operator=(const svlPointNd<V,N1>& p) {
	for ( int i=0; i<N && i<N1; ++i ) _vals[i] = p._vals[i];
	return *this;
}

template<class V, int N>
inline svlPointNd<V,N>& svlPointNd<V,N>::operator=(const V d) {
	for ( int i=0; i<N; ++i ) _vals[i] = d;
	return *this;
}

template<class V, int N>
inline bool svlPointNd<V,N>::operator>(const V d) {
	bool ret = true;
	for ( int i=0; ret && i<N; ++i ) ret = ret && _vals[i] > d;
	return ret;
}

template<class V, int N>
inline bool svlPointNd<V,N>::operator>=(const V d) {
	bool ret = true;
	for ( int i=0; ret && i<N; ++i ) ret = ret && _vals[i] >= d;
	return ret;
}

template<class V, int N>
inline bool svlPointNd<V,N>::operator<(const V d) {
	bool ret = true;
	for ( int i=0; ret && i<N; ++i ) ret = ret && _vals[i] < d;
	return ret;
}

template<class V, int N>
inline bool svlPointNd<V,N>::operator<=(const V d) {
	bool ret = true;
	for ( int i=0; ret && i<N; ++i ) ret = ret && _vals[i] <= d;
	return ret;
}

template<class V, int N>
template<int N1>
inline svlPointNd<V,N>& svlPointNd<V,N>::operator+=(const svlPointNd<V,N1>& p) {
	for ( int i=0; i<N && i<N1; ++i ) _vals[i] += p._vals[i];
	return *this;
}

template<class V, int N>
inline svlPointNd<V,N>& svlPointNd<V,N>::operator+=(const V d) {
	for ( int i=0; i<N; ++i ) _vals[i] += d;
	return *this;
}

template<class V, int N>
template<int N1>
inline svlPointNd<V,N>& svlPointNd<V,N>::operator-=(const svlPointNd<V,N1>& p) {
	for ( int i=0; i<N && i<N1; ++i ) _vals[i] -= p._vals[i];
	return *this;
}

template<class V, int N>
inline svlPointNd<V,N>& svlPointNd<V,N>::operator-=(const V d) {
	for ( int i=0; i<N; ++i ) _vals[i] -= d;
	return *this;
}

template<class V, int N>
template<int N1>
inline svlPointNd<V,N>& svlPointNd<V,N>::operator*=(const svlPointNd<V,N1>& p) {
	for ( int i=0; i<N && i<N1; ++i ) _vals[i] *= p._vals[i];
	return *this;
}

template<class V, int N>
inline svlPointNd<V,N>& svlPointNd<V,N>::operator*=(const V d) {
	for ( int i=0; i<N; ++i ) _vals[i] *= d;
	return *this;
}

template<class V, int N>
template<int N1>
inline svlPointNd<V,N>& svlPointNd<V,N>::operator/=(const svlPointNd<V,N1>& p) {
	for ( int i=0; i<N && i<N1; ++i ) _vals[i] /= p._vals[i];
	return *this;
}

template<class V, int N>
inline svlPointNd<V,N>& svlPointNd<V,N>::operator/=(const V d) {
	for ( int i=0; i<N; ++i ) _vals[i] /= d;
	return *this;
}

template<class V, int N>
template<int N1>
inline bool svlPointNd<V,N>::operator==(const svlPointNd<V,N1>& p) const {
	if ( N!=N1 ) return false;
	bool ret = true;
	for ( int i=0; ret && i<N; ++i ) ret = ret && (_vals[i] == p._vals[i]);
	return ret;
}

template<class V, int N>
inline bool svlPointNd<V,N>::operator==(const V d) const {
	bool ret = true;
	for ( int i=0; ret && i<N; ++i ) ret = ret && _vals[i] == d;
	return ret;
}

template<class V, int N>
template<int N1>
inline bool svlPointNd<V,N>::operator!=(const svlPointNd<V,N1>& p) const {
	return !(*this==p);
}

template<class V, int N>
inline bool svlPointNd<V,N>::operator!=(const V d) const {
	return !(*this==d);
}

// ****** Friend operators ******

template<class V, int N, class V1>
inline svlPointNd<V,N> operator+(const svlPointNd<V,N>& p, V1 d) {
	svlPointNd<V,N> ret;
	for ( int i=0; i<N; ++i ) ret._vals[i] = p._vals[i]+d;
	return ret;
}

template<class V, int N, class V1>
inline svlPointNd<V,N> operator+(V1 d, const svlPointNd<V,N>& p) {
	svlPointNd<V,N> ret;
	for ( int i=0; i<N; ++i ) ret._vals[i] = p._vals[i]+d;
	return ret;
}

template<class V, int N, class V1, int N1>
inline svlPointNd<V,N> operator+(const svlPointNd<V,N>& p, const svlPointNd<V1,N1>& q) {
	svlPointNd<V,N> ret;
	int i;
	for ( i=0; i<N && i<N1; ++i ) ret._vals[i] = p._vals[i] + q._vals[i];
	for ( ; i<N; ++i ) ret._vals[i] = p._vals[i];
	for ( ; i<N1; ++i ) ret._vals[i] = q._vals[i];
	return ret;
}

template<class V, int N, class V1>
inline svlPointNd<V,N> operator-(const svlPointNd<V,N>& p, V1 d) {
	svlPointNd<V,N> ret;
	for ( int i=0; i<N; ++i ) ret._vals[i] = p._vals[i]-(V)d;
	return ret;
}

template<class V, int N, class V1>
inline svlPointNd<V,N> operator-(V1 d, const svlPointNd<V,N>& p) {
	svlPointNd<V,N> ret;
	for ( int i=0; i<N; ++i ) ret._vals[i] = (V)d - p._vals[i];
	return ret;
}

template<class V, int N, class V1, int N1>
inline svlPointNd<V,N> operator-(const svlPointNd<V,N>& p, const svlPointNd<V1,N1>& q) {
	svlPointNd<V,N> ret;
	int i;
	for ( i=0; i<N && i<N1; ++i ) ret._vals[i] = p._vals[i] - (V)q._vals[i];
	for ( ; i<N; ++i ) ret._vals[i] = p._vals[i];
	for ( ; i<N1; ++i ) ret._vals[i] = -(V)q._vals[i];
	return ret;
}

template<class V, int N, class V1>
inline svlPointNd<V,N> operator*(V1 d, const svlPointNd<V,N>& p) {
	svlPointNd<V,N> ret;
	for ( int i=0; i<N; ++i ) ret._vals[i] = p._vals[i]*(V)d;
	return ret;
}

template<class V, int N, class V1>
inline svlPointNd<V,N> operator*(const svlPointNd<V,N>& p, V1 d) {
	svlPointNd<V,N> ret;
	for ( int i=0; i<N; ++i ) ret._vals[i] = p._vals[i]*(V)d;
	return ret;
}

template<class V, int N, class V1, int N1>
inline svlPointNd<V,N> operator*(const svlPointNd<V,N>& p, const svlPointNd<V1,N1>& q) {
	svlPointNd<V,N> ret;
	int i;
	for ( i=0; i<N && i<N1; ++i ) ret._vals[i] = (V)(p._vals[i]*q._vals[i]);
	for ( ; i<N; ++i ) ret._vals[i] = p._vals[i];
	return ret;
}

template<class V, int N, class V1>
inline svlPointNd<V,N> operator/(V1 d, const svlPointNd<V,N>& p) {
	svlPointNd<V,N> ret;
	for ( int i=0; i<N; ++i ) ret._vals[i] = (V)d / p._vals[i];
	return ret;
}

template<class V, int N, class V1>
inline svlPointNd<V,N> operator/(const svlPointNd<V,N>& p, V1 d) {
	svlPointNd<V,N> ret;
	for ( int i=0; i<N; ++i ) ret._vals[i] = p._vals[i]/(V)d;
	return ret;
}

template<class V, int N, class V1, int N1>
inline svlPointNd<V,N> operator/(const svlPointNd<V,N>& p, const svlPointNd<V1,N1>& q) {
	svlPointNd<V,N> ret;
	int i;
	for ( i=0; i<N && i<N1; ++i ) ret._vals[i] = p._vals[i]/(V)q._vals[i];
	for ( ; i<N; ++i ) ret._vals[i] = p._vals[i];
	return ret;
}

