/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Paul Baumstarck
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlTimer.h
** AUTHOR(S):   Paul Baumstarck <pbaumstarck@stanford.edu>
** DESCRIPTION:
**  Implements a compact, dynamic container for multiple timers based on time.h
**
*****************************************************************************/

#include <iostream>
#include <time.h>
using namespace std;

struct TimerUnit {
	clock_t time, elapsed;
	bool running;
	TimerUnit() : time(0), elapsed(0), running(false) {}
};

class svlTimer {
private:
	vector<TimerUnit> timers;

public:
	svlTimer(int n = 1) {
		timers = vector<TimerUnit>(n, TimerUnit());
	}
	virtual ~svlTimer() {}

	// Get the number of timers.
	inline size_t size() {
		return timers.size();
	}
	// Allocates just one timer, initialized to not running, and returns the index in the arrow.
	inline int alloc_timer() {
		timers.push_back(TimerUnit());
		return (int)timers.size()-1;
	}
	// Creates multiple timers initialized to not running.
	inline void alloc_timer(int n) {
		for ( int i=0; i<n; ++i )
			timers.push_back(TimerUnit());
	}
	// Starts a timer running.
	inline void start_timer(int i = 0) {
		timers[i].time = clock();
		timers[i].running = true;
	}
	// Sets the time elapsed to zero and begins counting again.
	inline void restart_timer(int i = 0) {
		timers[i].elapsed = 0;
		start_timer(i);
	}
	// Stops the timer running and tallies the elapsed time.
	inline clock_t stop_timer(int i = 0) {
		if ( timers[i].running ) {
			timers[i].elapsed += clock() - timers[i].time;
			timers[i].time = clock();
			timers[i].running = false;
		}
		return timers[i].elapsed;
	}
	// Gets the elapsed time from a timer.
	inline clock_t get_elapsed(int i = 0) {
		if ( timers[i].running ) {
			return timers[i].elapsed + (clock() - timers[i].time);
		}
		return timers[i].elapsed;
	}
	// Whether a timer is presently running or not.
	inline bool is_running(int i = 0) {
		return timers[i].running;
	}
	// Prints the elapsed time for a specific timer to a specified precision (default is milliseconds).
	inline void print(int i = 0, char d = 'm') {
		float mul = 1.0f;
		switch ( d ) {
			case 'm':
				mul = 1e3f; break;
			case 'u':
				mul = 1e6f; break;
			case 'n':
				mul = 1e9f; break;
		}
		printf("%0.3f %cs\n", float(get_elapsed(i)) * mul / float(CLOCKS_PER_SEC), d == 's' ? ' ' : d);
	}
	// Prints the elapsed time, as above, but divided by some number for a "time per X" value.
	inline void print_per(float per, int i = 0, char d = 'm') {
		float mul = 1.0f;
		switch ( d ) {
			case 'm':
				mul = 1e3f; break;
			case 'u':
				mul = 1e6f; break;
			case 'n':
				mul = 1e9f; break;
		}
		printf("%0.3f %cs per\n", float(get_elapsed(i)) * mul / float(CLOCKS_PER_SEC * per), d == 's' ? ' ' : d);
	}
};

