/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlFinalPatchDictionary.cpp
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Sid Batra <sidbatra@cs.stanford.edu>
**              David Breeden <breeden@cs.stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
**              Ian Goodfellow <ia3n@cs.stanford.edu>
**              Paul Baumstarck <pbaumstarck@stanford.edu>
**
*****************************************************************************/


#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>
#include <list>
#include <limits>
#include "cv.h"
#include "cxcore.h"
#include "highgui.h"

#include "svlBase.h"
#include "svlVision.h"
#include "svlDevel.h"

#if defined(_WIN32)||defined(WIN32)||defined(__WIN32__)
#undef max
#undef USE_THREADS
#endif

using namespace std;


// Initialize by reading from file.
svlFinalPatchDictionary::svlFinalPatchDictionary(const char *file)
{
	read(file);
}

// Initialize window size.
svlFinalPatchDictionary::svlFinalPatchDictionary(unsigned width, unsigned height)
{
	_windowSize = cvSize(width, height);
}

// Initialize from another dictionary.
svlFinalPatchDictionary::svlFinalPatchDictionary(const svlFinalPatchDictionary& d)
	: _intEntries(d._intEntries)
{
	_windowSize = d._windowSize;
}

svlFinalPatchDictionary::~svlFinalPatchDictionary()
{
	// Delete patches.
	for ( vector<svlIntensityPatch*>::iterator it=_intEntries.begin(); it!=_intEntries.end(); ++it )
		delete *it;
}

void svlFinalPatchDictionary::clear()
{
	for ( vector<svlIntensityPatch*>::iterator it=_intEntries.begin(); it!=_intEntries.end(); ++it )
		delete *it;
	_intEntries.clear();
}

// Ready dictionary entries out of a file.
bool svlFinalPatchDictionary::read(const char *filename)
{
	XMLNode root = XMLNode::parseFile(filename);
	bool ret = read(root);
	if ( !ret ) SVL_LOG(SVL_LOG_WARNING, "could not read patch dictionary file " << filename);
	return ret;
}

// Ready dictionary entries out of an XML tree.
bool svlFinalPatchDictionary::read(XMLNode &root)
{
	clear();
	if (root.isEmpty()) {
		SVL_LOG(SVL_LOG_WARNING, "Patch dictionary file is missing or has incorrect root.");
		return false;
	}
	if (string(root.getName()) != "PatchDictionary" ) {
		SVL_LOG(SVL_LOG_WARNING, "Attempt to read patch dictionary from XMLNode that is not a patch dictionary.");
		return false;
	}
	// Read window size.
	_windowSize.width = atoi(root.getAttribute("width"));
	_windowSize.height = atoi(root.getAttribute("height"));
	int n = atoi(root.getAttribute("numEntries"));

	assert(root.nChildNode("PatchDefinition") == n);

	// Read in patches.
	for ( int i=0; i<n; ++i ) {
		XMLNode node = root.getChildNode("PatchDefinition", i);
		const char *t = node.getAttribute("type");
		if ( !strcasecmp(t, "intensity") ) {
			_intEntries.push_back( svlIntensityPatchDefinition::readIntensityPatch(node) );
		} else if (!strcasecmp(t, "depth")) {
			cout << "slFinalPatchDictionary::read: Tried to read a depth patch.\n"; assert(false);
		} else if (!strcasecmp(t, "abs")) {
			cout << "slFinalPatchDictionary::read: Tried to read an abs patch.\n"; assert(false);
		} else {
			SVL_LOG(SVL_LOG_FATAL, "unknown patch type " << t);
		}

		assert(_entries[i] != NULL);
	}
	return true;
}

// Print all patches.
void svlFinalPatchDictionary::print()
{
	for ( vector<svlIntensityPatch*>::const_iterator it=_intEntries.begin(); it!=_intEntries.end(); ++it ) {
		printf("Patch:\n");
		(*it)->print();
	}
	printf("Dictionary window size (w,h): %d,%d\n", _windowSize.width, _windowSize.height);
}


