/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlPatchDefinition.h
** AUTHOR(S):   Stephen Gould <sgould@stanford.edu>
**              Sid Batra <sidbatra@cs.stanford.edu>
**              Olga Russakovsky <olga@cs.stanford.edu>
**              Ian Goodfellow <ia3n@cs.stanford.edu>
** DESCRIPTION:
**   Stores a generic definition of a patch. Which is inherited by patch
**   definitions to represent seperate channels. The patches are written
**   and read in XML format.
**
*****************************************************************************/

#pragma once

#include <cassert>
#include <iostream>
#include <vector>
#include "cv.h"
#include "cxcore.h"
#include "xmlParser/xmlParser.h"
#include "svlBase.h"



//Do not include svlDevel! It will make a cycle! You may include svlDevel from inside .cpp files only

using namespace std;

/* Very important global function to ensure that uniform patches
 * are not considered, and certainly never passed to cvMatchTemplate,
 * which would yield nan's and inf's */

bool isImageUniform(IplImage *img, double min_delta = 0);

bool isImageAlmostUniform(IplImage *img);



// svlPatchDefinitionType ----------------------------------------------------

typedef enum _svlPatchDefinitionType {
  SVL_INTENSITY_PATCH, // one-channel image patch
  SVL_DEPTH_PATCH,    // depthmap patch
  SVL_ABS_PATCH,      // works with the absolute differences between pairs of pixels in a template,
                      // and thus can be derived from greyscale, color, hyperspectral, whatever
                      // images
} svlPatchDefinitionType;

// svlPatchDefinition --------------------------------------------------------

class svlPatchDictionary;

class svlPatchDefinition 
{
    friend class svlPatchDictionary;

 protected:
    IplImage *_template;    // (single channel) patch template
    CvRect _validRect;       // valid response region
    int _validChannel;      // channel information

 protected:
    void setParent(svlPatchDictionary * parent) { _parent = parent; }

 public:
    svlPatchDefinition( svlPatchDictionary * parent );
    svlPatchDefinition(XMLNode& node, svlPatchDictionary * parent);
    svlPatchDefinition(const IplImage *t, const CvRect& rect, svlPatchDictionary * parent, int channel = 0);
    virtual ~svlPatchDefinition();

    // constuction helper functions
    static svlSmartPointer<svlPatchDefinition>
      createPatchDefinition(svlPatchDefinitionType t, svlPatchDictionary * parent);
    static svlSmartPointer<svlPatchDefinition>
      createPatchDefinition(XMLNode& node, svlPatchDictionary * parent);

    // i/o functions
    virtual svlPatchDefinitionType patchType() const = 0;
    virtual bool read(XMLNode& node);
    virtual bool write(ostream& os);

    //Returns the valid channel index
    int validChannel() const { return _validChannel; }
    void setValidChannel( int newChannel ) { _validChannel = newChannel ; }

    // Compute response image for input image. The caller is responsible for 
    // freeing memory. Input images should have the correct number of channels
    // for the patch feature. Non-vectorized inputs assume that the correct
    // channel is being passed in (usually channel 0).
    virtual IplImage *responseImage(const IplImage *image) const = 0;
    virtual IplImage *responseImage(const vector<IplImage *>& images) const;
       
    // Computes the response value for the patch at any given location of the
    // sliding window. responseImage should have been computed by responseImage()
    // function.
    virtual double patchValue(const IplImage *responseImage, const CvPoint& location) const;

    // Computes the response value for all the given patches; ideally this should
    // be optimized for sparsity, so should not just call responseImage followed
    // by locations.size() calls to patchValue
    virtual vector<double> patchValues(const IplImage *image, 
        const vector<CvPoint>& locations);
    virtual vector<double> patchValues(const vector<IplImage *>& images,
        const vector<CvPoint>& locations);

    virtual bool isPatchValid() = 0;

    bool isCorrelatedWith(svlPatchDefinition *patch, double threshold) const;

    // generally same channel is enough to guarantee that
    virtual bool isSameType(svlPatchDefinition *patch) const = 0;

	inline CvRect getValidRect() {
		return CvRect(_validRect);
	}
	inline int getTemplateWidth() {
		return _template->width;
	}
	inline int getTemplateHeight() {
		return _template->height;
	}

    //Implements the same functionality as a call to responseImage followed by patchValue(responseImage, cvPoint(0,0)), but classes that implement their own version of it can avoid the inefficiency of making the entire responseImage
    //warning-- will reset ROI of image
    virtual double trainingResponseValue(IplImage * image) const;


    svlPatchDictionary * _parent;
};

// Statistics on a patch.
struct svlPatchStats {
	float mean, std, *patch, *mask;
	int w, h;
	svlPatchStats() {}
	svlPatchStats(float mean, float std, float *patch, float *mask, int w, int h)
		: mean(mean), std(std), patch(patch), mask(mask), w(w), h(h) {}
	~svlPatchStats() {
		if ( patch ) delete patch;
		if ( mask ) delete mask;
	}
};

// svlIntensityPatchDefinition -----------------------------------------------

class svlIntensityPatchDefinition  : public svlPatchDefinition
{
 public:
    svlIntensityPatchDefinition(svlPatchDictionary * parent);
    svlIntensityPatchDefinition(XMLNode& node, svlPatchDictionary * parent);
    svlIntensityPatchDefinition(const IplImage *t, const CvRect& rect, svlPatchDictionary * parent, int channel = 0);
    ~svlIntensityPatchDefinition();
    
    svlPatchDefinitionType patchType() const { return SVL_INTENSITY_PATCH; }
	bool read(XMLNode &node);
	static svlIntensityPatch* readIntensityPatch(XMLNode &node);

    // response images
    IplImage *responseImage(const IplImage *image) const;
    IplImage *responseImage(const IplImage *image, const IplImage *imageSum,
        const IplImage *imageSumSq) const;
    IplImage *responseImage(vector<const IplImage *>& images,
        vector<const IplImage *> &imageSums,
        vector<const IplImage *> &imageSumSqs) const;
    
    vector<double> patchValues(const IplImage *image, 
        const vector<CvPoint>& locations);

    //warning-- will reset ROI of image
    double trainingResponseValue(IplImage * image) const;
    
    bool isPatchValid() { return !isImageUniform(_template); }
    bool isSameType(svlPatchDefinition *patch) const { return validChannel() == patch->validChannel(); }

protected:
	static map<IplImage*,svlPatchStats*> patch_cache;
	static void clearPatchCache() {
		for ( map<IplImage*,svlPatchStats*>::iterator it=patch_cache.begin(); it!=patch_cache.end(); ++it )
			delete it->second;
		patch_cache.clear();
	}
};

// svlDepthPatchDefinition ---------------------------------------------------

class svlDepthPatchDefinition  : public svlPatchDefinition
{
 protected:
    static int _maxDistance;
    static int _totalBins;
    int *_histogram;

 public:
    static CvSize _windowSize;

 public:
    svlDepthPatchDefinition(svlPatchDictionary * parent);
    svlDepthPatchDefinition(XMLNode& node, svlPatchDictionary * parent);
    svlDepthPatchDefinition(const IplImage *t, const CvRect& rect, svlPatchDictionary * parent, int channel = 1);
    ~svlDepthPatchDefinition();

    // i/o functions
    svlPatchDefinitionType patchType() const { return SVL_DEPTH_PATCH; }
    bool read(XMLNode& node);

    // response images
    IplImage *responseImage(const IplImage *image) const;

    vector<double> patchValues(const IplImage *image, const vector<CvPoint>& windows);

    bool isPatchValid() { return !isImageUniform(_template); }
    bool isSameType(svlPatchDefinition *patch) const { return validChannel() == patch->validChannel(); }

 protected:
    // takes the depth values of the template patch and computes a depth histogram for the selected region
    int** computeDepthHistogram(const IplImage* patchTemplate , CvRect region);
    void smoothHistogram(int *histogram, int totalLength);

    // simple euclidean distance between 2 vectors
    int computeEuclideanDistance(int *a , int *b , int totalLength);
    int computeEMD(int *a , int *b , int totalLength);

    // double patchValue(const IplImage *responseImage,  const CvPoint& location);
};


// svlAbsPatchDefinition ---------------------------------------------------

typedef enum _svlAbsPatchType {
  SVL_ABS_NE, SVL_ABS_E, SVL_ABS_SE, SVL_ABS_S, SVL_ABS_NUM,
  SVL_ABS_INIT, SVL_ABS_ONE, SVL_ABS_TWO, SVL_ABS_FOUR
} svlAbsPatchType;

class svlAbsPatchDefinition  : public svlPatchDefinition
{
 protected:
  CvMat **_masks; // floating-point
  int _nmasks;
  CvPoint *_refs;
  CvPoint _responseOffsetTL;
  CvPoint _responseOffsetBR;
  IplImage **_responses; // not the same as template because the matching will be done
                         // differently; depth == 8U
  svlAbsPatchType _type;
  //svlAbsPatchColorSpace _colorSpace;

  void svlAbsInitialize(svlAbsPatchType t);

 public:
  svlAbsPatchDefinition(svlPatchDictionary * parent, svlAbsPatchType t = (svlAbsPatchType)0);
  svlAbsPatchDefinition(XMLNode& node, svlPatchDictionary * parent);
  // this means that memory of masks and refs might be shared -- 
  // very careful with cleanup afterwards
  svlAbsPatchDefinition(const IplImage *t, const CvRect& rect, svlPatchDictionary * parent, int channel = 0,
			svlAbsPatchType tp = (svlAbsPatchType)0);
  //    svlAbsPatchDefinition(CvMat **m, CvPoint *r, int n, int ms)    
  ~svlAbsPatchDefinition();
  
  // minimum size of patches
  int getMinWidth() const { return _responseOffsetTL.x + _responseOffsetBR.x + 1; }
  int getMinHeight() const { return _responseOffsetTL.y + _responseOffsetBR.y + 1; }
  svlAbsPatchType getAbsType() const { return _type; }
  
  // i/o functions
  svlPatchDefinitionType patchType() const { return SVL_ABS_PATCH; }
  bool read(XMLNode& node);
  bool write(ostream& os);
  
  // Compute response image for input image. The caller is responsible for 
  // freeing memory. Input images should have the correct number of channels
  // for the patch feature. Non-vectorized inputs assume that the correct
  // channel is being passed in (usually channel 0).
  IplImage *responseImage(const IplImage *image) const;
  IplImage *responseImage(IplImage **imResponses) const;
    
  vector<double> patchValues(const IplImage *image, 
			     const vector<CvPoint>& locations);
  // for speedup, instead of passing in the image can pass in the resulting
  // convolutions with each of the _nmasks masks
  vector<double> patchValues(IplImage **imResponses, 
			     const vector<CvPoint>& locations);
  
  // Return the responses obtained by running each of the _nmasks
  // masks over the input image. In order to get the desired
  // response image we want to compare these responses to what we would
  // expect to get with our template patch.
  
  // the resulting images have depth 8U
  IplImage **maskResponseImages(const IplImage *image) const;
  void deallocateResponseImages(IplImage **&imResponses) const;

  bool isPatchValid() { 
    // if at least one of the response images is completely uniform,
    // then this is not a valid patch
    for (int i = 0; i < _nmasks; i++)
      if (isImageUniform(_responses[i])) return false;
    return true;
  }

  bool isSameType(svlPatchDefinition *patch) const {
    if (validChannel() != patch->validChannel() ||
	patch->patchType() != SVL_ABS_PATCH)
      return false;
    svlAbsPatchDefinition *p = (svlAbsPatchDefinition *)patch;
    return (_type == p->_type);
  }
};
